#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
import sys
import time
import argparse
from datetime import datetime, timedelta

from lsl.common.mcs import mjdmpm2datetime, mode2string
from lsl.common import metabundle, metabundleADP
from lsl.reader import tbw, tbn, drx, errors


def obsComp(x, y):
    """
    Function to help sort observations in time.
    """
    
    tX = mjdmpm2datetime(x['MJD'], x['MPM'])
    tY = mjdmpm2datetime(y['MJD'], y['MPM'])
    if tX < tY:
        return -1
    elif tX > tY:
        return 1
    else:
        return 0


def main(args):
    # Get the file names
    meta = args.metadata
    data = args.filename

    # Get all observations and their start times
    try:
        ## LWA-1
        sdf = metabundle.getSessionDefinition(meta)
        ses = metabundle.getSessionSpec(meta)
        obs = metabundle.getObservationSpec(meta)
    except:
        ## LWA-SV
        ### Try again
        sdf = metabundleADP.getSessionDefinition(meta)
        ses = metabundleADP.getSessionSpec(meta)
        obs = metabundleADP.getObservationSpec(meta)
    obs.sort(obsComp)
    tStart = []
    oDetails = []
    for i,o in enumerate(obs):
        tStart.append( mjdmpm2datetime(o['MJD'], o['MPM']) )
        oDetails.append( {'m': o['Mode'], 'd': o['Dur'] / 1000.0, 'f': o['BW'], 
                'p': o['projectID'], 's': o['sessionID'], 'o': o['obsID'], 't': sdf.sessions[0].observations[o['obsID']-1].target} )

        print "Observation #%i" % (o['obsID'])
        print " Start: %i, %i -> %s" % (o['MJD'], o['MPM'], tStart[-1])
        print " Mode: %s" % mode2string(o['Mode'])
        print " BW: %i" % o['BW']
        print " Target: %s" % sdf.sessions[0].observations[o['obsID']-1].target
    print " "

    # Figure out where in the file the various bits are.
    fh = open(data, 'rb')
    lf = drx.readFrame(fh)
    beam, j, k = lf.parseID()
    if beam != obs[0]['drxBeam']:
        print 'ERROR: Beam mis-match, metadata is for #%i, file is for #%i' % (obs[0]['drxBeam'], beam)
        sys.exit()
    firstFrame = datetime.utcfromtimestamp(lf.getTime())
    if abs(firstFrame - min(tStart)) > timedelta(seconds=30):
        print 'ERROR: Time mis-match, metadata is for %s, file is for %s' % (min(tStart), firstFrame)
        sys.exit()
    fh.seek(0)

    for i in xrange(len(tStart)):
        eof = False

        ## Get observation properties
        oStart = tStart[i]
        oMode = mode2string(oDetails[i]['m'])
        oDur  = oDetails[i]['d']
        oBW   = oDetails[i]['f']
        print "Seeking %s observation of %.3f seconds at %s" % (oMode, oDur, oStart)

        ## Get the correct reader to use
        if oMode == 'TBW':
            reader = tbw
            bwKey = None
            bwMult = 520.0 / 400
            fCount = 400
        elif oMode == 'TBN':
            reader = tbn
            bwKey = tbn.filterCodes
            bwMult = 520.0 / 512
            fCount = 512
        else:
            reader = drx
            bwKey = drx.filterCodes
            bwMult = 4.0 / 4096
            fCount = 4096

        ## Jump ahead to where the next frame should be, if needed
        if i != 0:
            pDur  = oDetails[i-1]['d']
            pBW   = oDetails[i-1]['f']

            nFramesSkip = int(pDur*bwKey[pBW]*bwMult)
            fh.seek(nFramesSkip*reader.FrameSize, 1)
            if fh.tell() >= os.path.getsize(data):
                fh.seek(-10*reader.FrameSize, 2)
                
        ## Figure out where we are and make sure we line up on a frame
        ## NOTE: This should never be needed
        fail = True
        while fail:
            try:
                frame = reader.readFrame(fh)
                fail = False
            except errors.syncError:
                fh.seek(1, 1)
            except errors.eofError:
                break
        fh.seek(-reader.FrameSize, 1)	

        ## Go in search of the start of the observation
        if datetime.utcfromtimestamp(frame.getTime()) < oStart:
            ### We aren't at the beginning yet, seek fowards
            print "-> At byte %i, time is %s < %s" % (fh.tell(), datetime.utcfromtimestamp(frame.getTime()), oStart)

            while datetime.utcfromtimestamp(frame.getTime()) < oStart:
                try:
                    frame = reader.readFrame(fh)
                except errors.syncError:		
                    fh.seek(1, 1)
                except errors.eofError:
                    break
                #print datetime.utcfromtimestamp(frame.getTime()), oStart

        elif datetime.utcfromtimestamp(frame.getTime()) > oStart:
            ### We've gone too far, seek backwards
            print "-> At byte %i, time is %s > %s" % (fh.tell(), datetime.utcfromtimestamp(frame.getTime()), oStart)

            while datetime.utcfromtimestamp(frame.getTime()) > oStart:
                if fh.tell() == 0:
                    break
                fh.seek(-2*reader.FrameSize, 1)
                try:
                    frame = reader.readFrame(fh)
                except errors.syncError:		
                    fh.seek(-1, 1)
                except errors.eofError:
                    break
                #print datetime.utcfromtimestamp(frame.getTime()), oStart
                
        else:
            ### We're there already
            print "-> At byte %i, time is %s = %s" % (fh.tell(), datetime.utcfromtimestamp(frame.getTime()), oStart)
            
        ## Jump back exactly one frame so that the filehandle is in a position 
        ## to read the first frame that is part of the observation
        try:
            frame = reader.readFrame(fh)
            print "-> At byte %i, time is %s = %s" % (fh.tell(), datetime.utcfromtimestamp(frame.getTime()), oStart)
            fh.seek(-reader.FrameSize, 1)
        except errors.eofError:
            pass
            
        ## Update the bytes ranges
        if fh.tell() < os.path.getsize(data):
            oDetails[i]['b'] = fh.tell()
            oDetails[i]['e'] = -1
        else:
            oDetails[i]['b'] = -1
            oDetails[i]['e'] = -1

        if i != 0:
            oDetails[i-1]['e'] = fh.tell()

        ## Progress report
        if oDetails[i]['b'] >= 0:
            print '-> Obs.', oDetails[i]['o'], 'starts at byte', oDetails[i]['b']
        else:
            print '-> Obs.', oDetails[i]['o'], 'starts after the end of the file'
    print " "

    # Report
    for i in xrange(len(tStart)):
        if oDetails[i]['b'] < 0:
            print "%s, Session %i, Observation %i: not found" % (oDetails[i]['p'], oDetails[i]['s'], oDetails[i]['o'])

        else:
            print "%s, Session %i, Observation %i: %i to %i (%i bytes)" % (oDetails[i]['p'], oDetails[i]['s'], oDetails[i]['o'], oDetails[i]['b'], oDetails[i]['e'], (oDetails[i]['e'] - oDetails[i]['b']))
    print " "

    # Split
    if not args.list:
        for i in xrange(len(tStart)):
            if oDetails[i]['b'] < 0:
                continue
                
            ## Report
            print "Working on Observation %i" % (i+1,)
            
            ## Create the output name
            if args.source:
                outname = '%s_%i_%s.dat' % (oDetails[i]['p'], oDetails[i]['s'], oDetails[i]['t'].replace(' ', '').replace('/','').replace('&','and'))
            else:
                outname = '%s_%i_%i.dat' % (oDetails[i]['p'], oDetails[i]['s'], oDetails[i]['o'])
                
            oMode = mode2string(oDetails[i]['m'])

            ## Get the correct reader to use
            if oMode == 'TBW':
                reader = tbw

            elif oMode == 'TBN':
                reader = tbn
            else:
                reader = drx

            ## Get the number of frames
            if oDetails[i]['e'] > 0:
                nFramesRead = (oDetails[i]['e'] - oDetails[i]['b']) / reader.FrameSize
            else:
                nFramesRead = (os.path.getsize(data) - oDetails[i]['b']) / reader.FrameSize

            ## Split
            if os.path.exists(outname):
                if not args.force:
                    yn = raw_input("WARNING: '%s' exists, overwrite? [Y/n] " % outname)
                else:
                    yn = 'y'
                    
                if yn not in ('n', 'N'):
                    os.unlink(outname)
                else:
                    print "WARNING: output file '%s' already exists, skipping" % outname
                    continue
                    
            fh.seek(oDetails[i]['b'])
            
            t0 = time.time()
            oh = open(outname, 'wb')
            for sl in [2**i for i in range(17)[::-1]]:
                while nFramesRead >= sl:
                    temp = fh.read(sl*reader.FrameSize)
                    oh.write(temp)
                    nFramesRead -= sl
            oh.close()
            t1 = time.time()
            print "  Copied %i bytes in %.3f s (%.3f MB/s)" % (os.path.getsize(outname), t1-t0, os.path.getsize(outname)/1024.0**2/(t1-t0))
    print " "


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description='given a MCS metadata tarball and a session data file, split the data file into individual observations', 
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
        )
    parser.add_argument('metadata', type=str, 
                        help='metadata file for the observation')
    parser.add_argument('filename', type=str, 
                        help='data file for the observation')
    parser.add_argument('-l', '--list', action='store_true', 
                        help='list source names')
    parser.add_argument('-s', '--source', action='store_true', 
                        help='split by source name instead of observation ID')
    parser.add_argument('-f', '--force', action='store_true', 
                        help='force overwritting of existing split files')
    args = parser.parse_args()

    main(args)
    



# -*- coding: utf-8 -*-

"""
lsl.misc - Miscellanous modules including:
 * mathutil       - math utilities that were part of the lwa_user package, 
 * geodesy        - module for retrieving earth orientation parameters,
 * beamformer     - post data aquisition beam former for TBW and TBN data,
 * ionosphere     - access to ionospheric and geomagnetic models, 
 * dedispersion   - dedispersion module,
 * scattering     - multi-path scattering module, 
 * autostereogram - make random dot autostereograms from 2-D NumPy arrays,
 * OrderedDict    - backport of the OrderedDict class from Python 2.7, 
 * rfutil         - convert RF engineering lingo into radio astrnomy lingo, and
 * wisdom         - build LSL-specific FFTW and PyFFTW wisdom.
"""





# -*- coding: utf-8 -*-

# Python3 compatiability
import sys
if sys.version_info > (3,):
    xrange = range
    
"""
Unit test for the lsl.common.sdf module.
"""

import os
import re
import pytz
import ephem
import tempfile
import unittest
from datetime import datetime, timedelta
try:
    import cStringIO as StringIO
except ImprotError:
    import StringIO

from lsl.common.paths import dataBuild as dataPath
from lsl.common import sdf
from lsl.common.stations import lwa1, lwasv


__revision__ = "$Rev: 2747 $"
__version__  = "0.4"
__author__    = "Jayce Dowell"


tbwFile = os.path.join(dataPath, 'tests', 'tbw-sdf.txt')
tbnFile = os.path.join(dataPath, 'tests', 'tbn-sdf.txt')
drxFile = os.path.join(dataPath, 'tests', 'drx-sdf.txt')
solFile = os.path.join(dataPath, 'tests', 'sol-sdf.txt')
jovFile = os.path.join(dataPath, 'tests', 'jov-sdf.txt')
stpFile = os.path.join(dataPath, 'tests', 'stp-sdf.txt')
spcFile = os.path.join(dataPath, 'tests', 'spc-sdf.txt')
tbfFile = os.path.join(dataPath, 'tests', 'tbf-sdf.txt')
idfFile = os.path.join(dataPath, 'tests', 'drx-idf.txt')


class sdf_tests(unittest.TestCase):
    """A unittest.TestCase collection of unit tests for the lsl.common.sdf
    module."""
    
    testPath = None

    def setUp(self):
        """Create the temporary file directory."""

        self.testPath = tempfile.mkdtemp(prefix='test-sdf-', suffix='.tmp')
        
    ### General ###
    
    def test_time(self):
        """Test the sdf.parseTime() function."""
        
        _UTC = pytz.utc
        _EST = pytz.timezone('US/Eastern')
        
        # Different realizations of the same thing
        s1 = "EST 2011-01-01 12:13:14.567"
        s2 = "EST 2011 01 01 12:13:14.567"
        s3 = "EST 2011 Jan 01 12:13:14.567"
        s4 = _EST.localize(datetime(2011, 1, 1, 12, 13, 14, 567000))
        s5 = _EST.localize(datetime(2011, 1, 1, 12, 13, 14, 567123))
        
        self.assertEqual(sdf.parseTime(s1), sdf.parseTime(s2))
        self.assertEqual(sdf.parseTime(s1), sdf.parseTime(s3))
        self.assertEqual(sdf.parseTime(s1), sdf.parseTime(s4))
        self.assertEqual(sdf.parseTime(s1), sdf.parseTime(s5))
        self.assertEqual(sdf.parseTime(s2), sdf.parseTime(s3))
        self.assertEqual(sdf.parseTime(s2), sdf.parseTime(s4))
        self.assertEqual(sdf.parseTime(s2), sdf.parseTime(s5))
        self.assertEqual(sdf.parseTime(s3), sdf.parseTime(s4))
        self.assertEqual(sdf.parseTime(s3), sdf.parseTime(s5))
        self.assertEqual(sdf.parseTime(s4), sdf.parseTime(s5))
        
        # Month name and month number agreement
        for n,m in enumerate(['Jan', 'Feb', 'Mar', 'Apr', 'May', 'Jun', 'Jul', 'Aug', 'Sep', 'Oct', 'Nov', 'Dec']):
            s1 = "UTC 2011-%s-14 12:13:14.000" % m
            s2 = "UTC 2011-%02i-14 12:13:14.000" % (n+1)
            s3 = _UTC.localize(datetime(2011, n+1, 14, 12, 13, 14, 0))
            self.assertEqual(sdf.parseTime(s1), sdf.parseTime(s2))
            self.assertEqual(sdf.parseTime(s1), sdf.parseTime(s3))
            
        # Time zone agreement - UTC
        s1 = "2011-01-01 12:13:14.567"
        s2 = "2011 01 01 12:13:14.567"
        s3 = _UTC.localize(datetime(2011, 1, 1, 12, 13, 14, 567000))
        self.assertEqual(sdf.parseTime(s1), sdf.parseTime(s2))
        self.assertEqual(sdf.parseTime(s1), sdf.parseTime(s3))
        self.assertEqual(sdf.parseTime(s2), sdf.parseTime(s3))
        
        # Time zone agreement - local
        for o,z in enumerate(['EST', 'CST', 'MST', 'PST']):
            h = 12
            o = -5 - o
            s1 = "%s 2011 01 01 %02i:13:14.567" % ('UTC', h)
            s2 = "%s 2011 01 01 %02i:13:14.567" % (z, h+o)
            self.assertEqual(sdf.parseTime(s1), sdf.parseTime(s2))
            
        # Something strange
        s1 = "CET 2013-01-08 19:42:00.000"
        s2 = "2013-01-08 18:42:00.000+00:00"
        s3 = "2013-01-08 11:42:00-0700"
        self.assertEqual(sdf.parseTime(s1), sdf.parseTime(s2))
        self.assertEqual(sdf.parseTime(s1), sdf.parseTime(s3))
        self.assertEqual(sdf.parseTime(s2), sdf.parseTime(s3))
        
        # Details
        s1 = "2011 01 02 03:04:05.678"
        out = sdf.parseTime(s1)
        ## Date
        self.assertEqual(out.year, 2011)
        self.assertEqual(out.month, 1)
        self.assertEqual(out.day, 2)
        ## Time
        self.assertEqual(out.hour, 3)
        self.assertEqual(out.minute, 4)
        self.assertEqual(out.second, 5)
        self.assertEqual(out.microsecond, 678000)
        
        # LST at LWA1
        s1 = "LST 2013-01-08 19:42:00.000"
        s2 = "UTC 2013-01-08 19:38:26.723"
        self.assertEqual(sdf.parseTime(s1, station=lwa1), sdf.parseTime(s2))
        
    def test_flat_projects(self):
        """Test single session/observations SDFs."""
        
        obs = sdf.Observer('Test Observer', 99)
        targ = sdf.DRX('Target', 'Target', '2019/1/1 00:00:00', '00:00:10', 0.0, 90.0, 40e6, 50e6, 7, MaxSNR=False)
        sess = sdf.Session('Test Session', 1, observations=targ)
        sess.setDRXBeam(1)
        proj = sdf.Project(obs, 'Test Project', 'COMTST', sessions=sess)
        out = proj.render()
        
    def test_ucf_username(self):
        """Test setting the UCF username for auto-copy support."""
        
        obs = sdf.Observer('Test Observer', 99)
        targ = sdf.DRX('Target', 'Target', '2019/1/1 00:00:00', '00:00:10', 0.0, 90.0, 40e6, 50e6, 7, MaxSNR=False)
        sess = sdf.Session('Test Session', 1, observations=targ)
        sess.setDRXBeam(1)
        sess.setDataReturnMethod('UCF')
        sess.setUCFUsername('test')
        proj = sdf.Project(obs, 'Test Project', 'COMTST', sessions=sess)
        out = proj.render()
        self.assertTrue(out.find('ucfuser:test') >= 0)
        
        obs = sdf.Observer('Test Observer', 99)
        targ = sdf.DRX('Target', 'Target', '2019/1/1 00:00:00', '00:00:10', 0.0, 90.0, 40e6, 50e6, 7, MaxSNR=False)
        sess = sdf.Session('Test Session', 1, observations=targ, comments='This is a comment')
        sess.setDRXBeam(1)
        sess.setDataReturnMethod('UCF')
        sess.setUCFUsername('test/dir1')
        proj = sdf.Project(obs, 'Test Project', 'COMTST', sessions=sess)
        out = proj.render()
        self.assertTrue(out.find('ucfuser:test/dir1') >= 0)
        
    ### TBW ###
    
    def test_tbw_parse(self):
        """Test reading in a TBW SDF file."""
        
        project = sdf.parseSDF(tbwFile)
        
        # Basic file structure
        self.assertEqual(len(project.sessions), 1)
        self.assertEqual(len(project.sessions[0].observations), 2)
        
        # Observational setup - 1
        self.assertEqual(project.sessions[0].observations[0].mode,  'TBW')
        self.assertEqual(project.sessions[0].observations[0].mjd,   55616)
        self.assertEqual(project.sessions[0].observations[0].mpm,       0)
        
        # Observational setup - 2
        self.assertEqual(project.sessions[0].observations[1].mode,  'TBW')
        self.assertEqual(project.sessions[0].observations[1].mjd,   55616)
        self.assertEqual(project.sessions[0].observations[1].mpm,  700000)
        
    def test_tbw_update(self):
        """Test updating TRK_SOL values."""
        
        project = sdf.parseSDF(tbwFile)
        project.sessions[0].observations[1].setStart("MST 2011 Feb 23 17:10:15")
        
        self.assertEqual(project.sessions[0].observations[1].mjd,  55616)
        self.assertEqual(project.sessions[0].observations[1].mpm,  615000)
        
    def test_tbw_write(self):
        """Test writing a TBW SDF file."""
        
        project = sdf.parseSDF(tbwFile)
        out = project.render()
        
    def test_tbw_errors(self):
        """Test various TBW SDF errors."""
        
        project = sdf.parseSDF(tbwFile)
        
        # Bad number of TBW bits
        project.sessions[0].observations[0].bits = 6
        self.assertFalse(project.validate())
        
        # Bad number of TBW samples
        project.sessions[0].observations[0].bits = 4
        project.sessions[0].observations[0].samples = 72000000
        self.assertFalse(project.validate())
        
        project.sessions[0].observations[0].bits = 12
        project.sessions[0].observations[0].samples = 72000000
        self.assertFalse(project.validate())
        
    ### TBN ###
    
    def test_tbn_parse(self):
        """Test reading in a TBN SDF file."""
        
        project = sdf.parseSDF(tbnFile)
        
        # Basic file structure
        self.assertEqual(len(project.sessions), 1)
        self.assertEqual(len(project.sessions[0].observations), 2)
        
        # Observational setup - 1
        self.assertEqual(project.sessions[0].observations[0].mode, 'TBN')
        self.assertEqual(project.sessions[0].observations[0].mjd,  55616)
        self.assertEqual(project.sessions[0].observations[0].mpm,      0)
        self.assertEqual(project.sessions[0].observations[0].dur,  10000)
        self.assertEqual(project.sessions[0].observations[0].freq1, 438261968)
        self.assertEqual(project.sessions[0].observations[0].filter,   7)
        
        # Observational setup - 2
        self.assertEqual(project.sessions[0].observations[1].mode, 'TBN')
        self.assertEqual(project.sessions[0].observations[1].mjd,  55616)
        self.assertEqual(project.sessions[0].observations[1].mpm,  10000)
        self.assertEqual(project.sessions[0].observations[1].dur,  10000)
        self.assertEqual(project.sessions[0].observations[1].freq1, 832697741)
        self.assertEqual(project.sessions[0].observations[1].filter,   7)
        
    def test_tbn_update(self):
        """Test updating TBN values."""
        
        project = sdf.parseSDF(tbnFile)
        project.sessions[0].observations[1].setStart("MST 2011 Feb 23 17:00:15")
        project.sessions[0].observations[1].setDuration(timedelta(seconds=15))
        project.sessions[0].observations[1].setFrequency1(75e6)
        
        self.assertEqual(project.sessions[0].observations[1].mjd,  55616)
        self.assertEqual(project.sessions[0].observations[1].mpm,  15000)
        self.assertEqual(project.sessions[0].observations[1].dur,  15000)
        self.assertEqual(project.sessions[0].observations[1].freq1, 1643482384)
        
    def test_tbn_write(self):
        """Test writing a TBN SDF file."""
        
        project = sdf.parseSDF(tbnFile)
        out = project.render()
        
    def test_tbn_errors(self):
        """Test various TBN SDF errors."""
        
        project = sdf.parseSDF(tbnFile)
        
        # Bad filter
        project.sessions[0].observations[0].filter = 10
        self.assertFalse(project.validate())
        
        # Bad frequency
        project.sessions[0].observations[0].filter = 7
        project.sessions[0].observations[0].frequency1 = 95.0e6
        project.sessions[0].observations[0].update()
        self.assertFalse(project.validate())
        
        # Bad duration
        project.sessions[0].observations[0].frequency1 = 38.0e6
        project.sessions[0].observations[0].duration = '96:00:00.000'
        project.sessions[0].observations[0].update()
        self.assertFalse(project.validate())
        
    ### DRX - TRK_RADEC ###
    
    def test_drx_parse(self):
        """Test reading in a TRK_RADEC SDF file."""
        
        project = sdf.parseSDF(drxFile)
        
        # Basic file structure
        self.assertEqual(len(project.sessions), 1)
        self.assertEqual(len(project.sessions[0].observations), 2)
        
        # Observational setup - 1
        self.assertEqual(project.sessions[0].observations[0].mode, 'TRK_RADEC')
        self.assertEqual(project.sessions[0].observations[0].mjd,  55616)
        self.assertEqual(project.sessions[0].observations[0].mpm,      0)
        self.assertEqual(project.sessions[0].observations[0].dur,  10000)
        self.assertEqual(project.sessions[0].observations[0].freq1,  438261968)
        self.assertEqual(project.sessions[0].observations[0].freq2, 1928352663)
        self.assertEqual(project.sessions[0].observations[0].filter,   7)
        self.assertAlmostEqual(project.sessions[0].observations[0].ra, 5.6, 6)
        self.assertAlmostEqual(project.sessions[0].observations[0].dec, 22.0, 6)
        
        # Observational setup - 2
        self.assertEqual(project.sessions[0].observations[1].mode, 'TRK_RADEC')
        self.assertEqual(project.sessions[0].observations[1].mjd,  55616)
        self.assertEqual(project.sessions[0].observations[1].mpm,  10000)
        self.assertEqual(project.sessions[0].observations[1].dur,  10000)
        self.assertEqual(project.sessions[0].observations[1].freq1,  832697741)
        self.assertEqual(project.sessions[0].observations[1].freq2, 1621569285)
        self.assertEqual(project.sessions[0].observations[1].filter,   7)
        self.assertAlmostEqual(project.sessions[0].observations[1].ra, 5.6, 6)
        self.assertAlmostEqual(project.sessions[0].observations[1].dec, 22.0, 6)
        
    def test_drx_update(self):
        """Test updating TRK_RADEC values."""
        
        project = sdf.parseSDF(drxFile)
        project.sessions[0].observations[1].setStart("MST 2011 Feb 23 17:00:15")
        project.sessions[0].observations[1].setDuration(timedelta(seconds=15))
        project.sessions[0].observations[1].setFrequency1(75e6)
        project.sessions[0].observations[1].setFrequency2(76e6)
        project.sessions[0].observations[1].setRA(ephem.hours('5:30:00'))
        project.sessions[0].observations[1].setDec(ephem.degrees('+22:30:00'))
        
        self.assertEqual(project.sessions[0].observations[1].mjd,  55616)
        self.assertEqual(project.sessions[0].observations[1].mpm,  15000)
        self.assertEqual(project.sessions[0].observations[1].dur,  15000)
        self.assertEqual(project.sessions[0].observations[1].freq1, 1643482384)
        self.assertEqual(project.sessions[0].observations[1].freq2, 1665395482)
        self.assertAlmostEqual(project.sessions[0].observations[1].ra, 5.5, 6)
        self.assertAlmostEqual(project.sessions[0].observations[1].dec, 22.5, 6)
        
    def test_drx_write(self):
        """Test writing a TRK_RADEC SDF file."""
        
        project = sdf.parseSDF(drxFile)
        out = project.render()
        
    def test_drx_errors(self):
        """Test various TRK_RADEC SDF errors."""
        
        project = sdf.parseSDF(drxFile)
        
        # Bad beam
        project.sessions[0].drxBeam = 6
        self.assertFalse(project.validate())
        
        # No beam
        project.sessions[0].drxBeam = -1
        self.assertFalse(project.validate())
        
        # Bad filter
        project.sessions[0].observations[0].filter = 10
        self.assertFalse(project.validate())
        
        # Bad frequency
        project.sessions[0].observations[0].filter = 7
        project.sessions[0].observations[0].frequency1 = 90.0e6
        project.sessions[0].observations[0].update()
        self.assertFalse(project.validate())
        
        project.sessions[0].observations[0].frequency1 = 38.0e6
        project.sessions[0].observations[0].frequency2 = 90.0e6
        project.sessions[0].observations[0].update()
        self.assertFalse(project.validate())
        
        # Bad duration
        project.sessions[0].observations[0].frequency2 = 38.0e6
        project.sessions[0].observations[0].duration = '96:00:00.000'
        project.sessions[0].observations[0].update()
        self.assertFalse(project.validate())
        
        # Bad pointing
        project.sessions[0].observations[0].duration = '00:00:01.000'
        project.sessions[0].observations[0].dec = -72.0
        project.sessions[0].observations[0].update()
        self.assertFalse(project.validate())
        
    ### DRX - TRK_SOL ###
    
    def test_sol_parse(self):
        """Test reading in a TRK_SOL SDF file."""
        
        project = sdf.parseSDF(solFile)
        
        # Basic file structure
        self.assertEqual(len(project.sessions), 1)
        self.assertEqual(len(project.sessions[0].observations), 2)
        
        # Observational setup - 1
        self.assertEqual(project.sessions[0].observations[0].mode, 'TRK_SOL')
        self.assertEqual(project.sessions[0].observations[0].mjd,  55616)
        self.assertEqual(project.sessions[0].observations[0].mpm,      0)
        self.assertEqual(project.sessions[0].observations[0].dur,  10000)
        self.assertEqual(project.sessions[0].observations[0].freq1,  438261968)
        self.assertEqual(project.sessions[0].observations[0].freq2, 1928352663)
        self.assertEqual(project.sessions[0].observations[0].filter,   7)
        
        # Observational setup - 2
        self.assertEqual(project.sessions[0].observations[1].mode, 'TRK_SOL')
        self.assertEqual(project.sessions[0].observations[1].mjd,  55616)
        self.assertEqual(project.sessions[0].observations[1].mpm,  10000)
        self.assertEqual(project.sessions[0].observations[1].dur,  10000)
        self.assertEqual(project.sessions[0].observations[1].freq1,  832697741)
        self.assertEqual(project.sessions[0].observations[1].freq2, 1621569285)
        self.assertEqual(project.sessions[0].observations[1].filter,   7)
        
    def test_sol_update(self):
        """Test updating TRK_SOL values."""
        
        project = sdf.parseSDF(solFile)
        project.sessions[0].observations[1].setStart("MST 2011 Feb 23 17:00:15")
        project.sessions[0].observations[1].setDuration(timedelta(seconds=15))
        project.sessions[0].observations[1].setFrequency1(75e6)
        project.sessions[0].observations[1].setFrequency2(76e6)
        
        self.assertEqual(project.sessions[0].observations[1].mjd,  55616)
        self.assertEqual(project.sessions[0].observations[1].mpm,  15000)
        self.assertEqual(project.sessions[0].observations[1].dur,  15000)
        self.assertEqual(project.sessions[0].observations[1].freq1, 1643482384)
        self.assertEqual(project.sessions[0].observations[1].freq2, 1665395482)
        
    def test_sol_write(self):
        """Test writing a TRK_SOL SDF file."""
        
        project = sdf.parseSDF(solFile)
        out = project.render()
        
    def test_sol_errors(self):
        """Test various TRK_SOL SDF errors."""
        
        project = sdf.parseSDF(solFile)
        
        # Bad beam
        project.sessions[0].drxBeam = 6
        self.assertFalse(project.validate())
        
        # No beam
        project.sessions[0].drxBeam = -1
        self.assertFalse(project.validate())
        
        # Bad filter
        project.sessions[0].observations[0].filter = 10
        self.assertFalse(project.validate())
        
        # Bad frequency
        project.sessions[0].observations[0].filter = 7
        project.sessions[0].observations[0].frequency1 = 90.0e6
        project.sessions[0].observations[0].update()
        self.assertFalse(project.validate())
        
        project.sessions[0].observations[0].frequency1 = 38.0e6
        project.sessions[0].observations[0].frequency2 = 90.0e6
        project.sessions[0].observations[0].update()
        self.assertFalse(project.validate())
        
        # Bad duration
        project.sessions[0].observations[0].frequency2 = 38.0e6
        project.sessions[0].observations[0].duration = '96:00:00.000'
        project.sessions[0].observations[0].update()
        self.assertFalse(project.validate())
        
    ### DRX - TRK_JOV ###
    
    def test_jov_parse(self):
        """Test reading in a TRK_JOV SDF file."""
        
        project = sdf.parseSDF(jovFile)
        
        # Basic file structure
        self.assertEqual(len(project.sessions), 1)
        self.assertEqual(len(project.sessions[0].observations), 2)
        
        # Observational setup - 1
        self.assertEqual(project.sessions[0].observations[0].mode, 'TRK_JOV')
        self.assertEqual(project.sessions[0].observations[0].mjd,  55616)
        self.assertEqual(project.sessions[0].observations[0].mpm,      0)
        self.assertEqual(project.sessions[0].observations[0].dur,  10000)
        self.assertEqual(project.sessions[0].observations[0].freq1,  438261968)
        self.assertEqual(project.sessions[0].observations[0].freq2, 1928352663)
        self.assertEqual(project.sessions[0].observations[0].filter,   7)
        
        # Observational setup - 2
        self.assertEqual(project.sessions[0].observations[1].mode, 'TRK_JOV')
        self.assertEqual(project.sessions[0].observations[1].mjd,  55616)
        self.assertEqual(project.sessions[0].observations[1].mpm,  10000)
        self.assertEqual(project.sessions[0].observations[1].dur,  10000)
        self.assertEqual(project.sessions[0].observations[1].freq1,  832697741)
        self.assertEqual(project.sessions[0].observations[1].freq2, 1621569285)
        self.assertEqual(project.sessions[0].observations[1].filter,   7)
        
    def test_jov_update(self):
        """Test updating TRK_JOV values."""
        
        project = sdf.parseSDF(jovFile)
        project.sessions[0].observations[1].setStart("MST 2011 Feb 23 17:00:15")
        project.sessions[0].observations[1].setDuration(timedelta(seconds=15))
        project.sessions[0].observations[1].setFrequency1(75e6)
        project.sessions[0].observations[1].setFrequency2(76e6)
        
        self.assertEqual(project.sessions[0].observations[1].mjd,  55616)
        self.assertEqual(project.sessions[0].observations[1].mpm,  15000)
        self.assertEqual(project.sessions[0].observations[1].dur,  15000)
        self.assertEqual(project.sessions[0].observations[1].freq1, 1643482384)
        self.assertEqual(project.sessions[0].observations[1].freq2, 1665395482)
        
    def test_jov_write(self):
        """Test writing a TRK_JOV SDF file."""
        
        project = sdf.parseSDF(jovFile)
        out = project.render()
        
    def test_jov_errors(self):
        """Test various TRK_JOV SDF errors."""
        
        project = sdf.parseSDF(jovFile)
        
        # Bad beam
        project.sessions[0].drxBeam = 6
        self.assertFalse(project.validate())
        
        # No beam
        project.sessions[0].drxBeam = -1
        self.assertFalse(project.validate())
        
        # Bad filter
        project.sessions[0].observations[0].filter = 10
        self.assertFalse(project.validate())
        
        # Bad frequency
        project.sessions[0].observations[0].filter = 7
        project.sessions[0].observations[0].frequency1 = 90.0e6
        project.sessions[0].observations[0].update()
        self.assertFalse(project.validate())
        
        project.sessions[0].observations[0].frequency1 = 38.0e6
        project.sessions[0].observations[0].frequency2 = 90.0e6
        project.sessions[0].observations[0].update()
        self.assertFalse(project.validate())
        
        # Bad duration
        project.sessions[0].observations[0].frequency2 = 38.0e6
        project.sessions[0].observations[0].duration = '96:00:00.000'
        project.sessions[0].observations[0].update()
        self.assertFalse(project.validate())
        
    ### DRX - STEPPED ###
    
    def test_stp_parse(self):
        """Test reading in a STEPPED SDF file."""
        
        project = sdf.parseSDF(stpFile)
        
        # Basic file structure
        self.assertEqual(len(project.sessions), 1)
        self.assertEqual(len(project.sessions[0].observations), 2)
        
        # Observational setup - 1
        self.assertEqual(project.sessions[0].observations[0].mode, 'STEPPED')
        self.assertEqual(project.sessions[0].observations[0].mjd,  55616)
        self.assertEqual(project.sessions[0].observations[0].mpm, 440000)
        self.assertEqual(project.sessions[0].observations[0].dur, 300000)
        self.assertEqual(project.sessions[0].observations[0].filter,   7)
        self.assertEqual(project.sessions[0].observations[0].obsFEE[0], [1,1])
        self.assertEqual(project.sessions[0].observations[0].aspFlt[0], 2)
        self.assertEqual(project.sessions[0].observations[0].aspAT1[0], 10)
        self.assertEqual(project.sessions[0].observations[0].aspAT2[0], 12)
        self.assertEqual(project.sessions[0].observations[0].aspATS[0], 14)
        
        # Steps - 1
        self.assertEqual(len(project.sessions[0].observations[0].steps), 4)
        for i in xrange(4):
            self.assertEqual(project.sessions[0].observations[0].steps[i].RADec, project.sessions[0].observations[0].RADec)
            self.assertEqual(project.sessions[0].observations[0].steps[i].freq1,  832697741)
            self.assertEqual(project.sessions[0].observations[0].steps[i].freq2, 1621569285)
        self.assertAlmostEqual(project.sessions[0].observations[0].steps[0].c1, 90.0, 6)
        self.assertAlmostEqual(project.sessions[0].observations[0].steps[0].c2, 45.0, 6)
        self.assertEqual(project.sessions[0].observations[0].steps[0].dur, 60000)
        self.assertAlmostEqual(project.sessions[0].observations[0].steps[-1].c1, 0.0, 6)
        self.assertAlmostEqual(project.sessions[0].observations[0].steps[-1].c2, 1.0, 6)
        self.assertEqual(project.sessions[0].observations[0].steps[-1].dur, 120000)
        
        # Observational setup - 2
        self.assertEqual(project.sessions[0].observations[1].mode, 'STEPPED')
        self.assertEqual(project.sessions[0].observations[1].mjd,  55616)
        self.assertEqual(project.sessions[0].observations[1].mpm, 800000)
        self.assertEqual(project.sessions[0].observations[1].dur, 180000)
        self.assertEqual(project.sessions[0].observations[1].filter,   7)
        self.assertEqual(project.sessions[0].observations[1].obsFEE[0], [1,0])
        self.assertEqual(project.sessions[0].observations[1].aspFlt[0], 1)
        self.assertEqual(project.sessions[0].observations[1].aspAT1[0], 11)
        self.assertEqual(project.sessions[0].observations[1].aspAT2[0], 13)
        self.assertEqual(project.sessions[0].observations[1].aspATS[0], 15)
        
        # Steps - 2
        self.assertEqual(len(project.sessions[0].observations[1].steps), 2)
        for i in xrange(2):
            self.assertEqual(project.sessions[0].observations[1].steps[i].RADec, project.sessions[0].observations[1].RADec)
            self.assertEqual(project.sessions[0].observations[1].steps[i].freq1,  832697741)
            self.assertEqual(project.sessions[0].observations[1].steps[i].freq2, 1621569285)
        self.assertAlmostEqual(project.sessions[0].observations[1].steps[0].c1, 0.0, 6)
        self.assertAlmostEqual(project.sessions[0].observations[1].steps[0].c2, 90.0, 6)
        self.assertEqual(project.sessions[0].observations[1].steps[0].dur, 60000)
        self.assertAlmostEqual(project.sessions[0].observations[1].steps[-1].c1, 12.0, 6)
        self.assertAlmostEqual(project.sessions[0].observations[1].steps[-1].c2, 80.0, 6)
        self.assertEqual(project.sessions[0].observations[1].steps[-1].dur, 120000)
        
    def test_stp_update(self):
        """Test updating a STEPPED SDF file."""
        
        project = sdf.parseSDF(stpFile)
        project.sessions[0].observations[1].setStart("MST 2011 Feb 23 17:00:15")
        for step in project.sessions[0].observations[1].steps:
            step.setDuration(timedelta(seconds=15))
            step.setFrequency1(75e6)
            step.setFrequency2(76e6)
            step.setC1(ephem.hours('10:30:00'))
            step.setC2(ephem.degrees('89:30:00'))
        project.sessions[0].observations[1].update()
        
        self.assertEqual(project.sessions[0].observations[1].mjd,  55616)
        self.assertEqual(project.sessions[0].observations[1].mpm,  15000)
        self.assertEqual(project.sessions[0].observations[1].dur,  30000)
        for step in project.sessions[0].observations[1].steps:
            self.assertEqual(step.dur, 15000)
            self.assertEqual(step.freq1, 1643482384)
            self.assertEqual(step.freq2, 1665395482)
            self.assertEqual(step.c1, 10.5)
            self.assertEqual(step.c2, 89.5)
            
        project = sdf.parseSDF(stpFile)
        project.sessions[0].observations[1].RADec = False
        project.sessions[0].observations[1].setStart("MST 2011 Feb 23 17:00:15")
        for step in project.sessions[0].observations[1].steps:
            step.RADec = False
            step.setDuration(timedelta(seconds=15))
            step.setFrequency1(75e6)
            step.setFrequency2(76e6)
            step.setC1(ephem.hours('10:30:00'))
            step.setC2(ephem.degrees('89:30:00'))
        project.sessions[0].observations[1].update()
        
        self.assertEqual(project.sessions[0].observations[1].mjd,  55616)
        self.assertEqual(project.sessions[0].observations[1].mpm,  15000)
        self.assertEqual(project.sessions[0].observations[1].dur,  30000)
        for step in project.sessions[0].observations[1].steps:
            self.assertFalse(step.RADec)
            self.assertEqual(step.dur, 15000)
            self.assertEqual(step.freq1, 1643482384)
            self.assertEqual(step.freq2, 1665395482)
            self.assertEqual(step.c1, 10.5*15)
            self.assertEqual(step.c2, 89.5)
        
    def test_stp_write(self):
        """Test writing a STEPPED SDF file."""
        
        project = sdf.parseSDF(stpFile)
        out = project.render()
        
    def test_stp_errors(self):
        """Test various STEPPED SDF errors."""
        
        project = sdf.parseSDF(stpFile)
        
        # Bad beam
        project.sessions[0].drxBeam = 6
        self.assertFalse(project.validate())
        
        # No beam
        project.sessions[0].drxBeam = -1
        self.assertFalse(project.validate())
        
        # Bad filter
        project.sessions[0].observations[0].filter = 10
        self.assertFalse(project.validate())
        
        # Bad frequency
        project.sessions[0].observations[0].filter = 7
        project.sessions[0].observations[0].steps[0].frequency1 = 90.0e6
        project.sessions[0].observations[0].update()
        self.assertFalse(project.validate())
        
        project.sessions[0].observations[0].steps[0].frequency1 = 38.0e6
        project.sessions[0].observations[0].steps[1].frequency2 = 90.0e6
        project.sessions[0].observations[0].update()
        self.assertFalse(project.validate())
        
        # Bad duration
        project.sessions[0].observations[0].steps[1].frequency2 = 38.0e6
        project.sessions[0].observations[0].steps[2].duration = '96:00:00.000'
        project.sessions[0].observations[0].update()
        self.assertFalse(project.validate())
        
    ### DRX - STEPPED with delays and gains ###
    
    def test_spc_parse(self):
        """Test reading in a STEPPED Delay and Gain SDF file."""
        
        project = sdf.parseSDF(spcFile)
        
        # Basic file structure
        self.assertEqual(len(project.sessions), 1)
        self.assertEqual(len(project.sessions[0].observations), 1)
        
        # Observational setup - 1
        self.assertEqual(project.sessions[0].observations[0].mode, 'STEPPED')
        self.assertEqual(project.sessions[0].observations[0].mjd,  55616)
        self.assertEqual(project.sessions[0].observations[0].mpm, 440000)
        self.assertEqual(project.sessions[0].observations[0].dur,  60000)
        self.assertEqual(project.sessions[0].observations[0].filter,   7)
        
        # Steps - 1
        self.assertEqual(len(project.sessions[0].observations[0].steps), 1)
        self.assertEqual(project.sessions[0].observations[0].steps[0].RADec, project.sessions[0].observations[0].RADec)
        self.assertAlmostEqual(project.sessions[0].observations[0].steps[0].c1, 90.0, 6)
        self.assertAlmostEqual(project.sessions[0].observations[0].steps[0].c2, 45.0, 6)
        self.assertEqual(project.sessions[0].observations[0].steps[0].freq1,  832697741)
        self.assertEqual(project.sessions[0].observations[0].steps[0].freq2, 1621569285)
        self.assertEqual(project.sessions[0].observations[0].steps[0].dur, 60000)
        
        # Delays - 1
        for i in xrange(260):
            self.assertEqual(project.sessions[0].observations[0].steps[0].delays[i], 0)
            
        # Gains - 1
        for i in xrange(260):
            self.assertEqual(project.sessions[0].observations[0].steps[0].gains[i][0][0], 1)
            self.assertEqual(project.sessions[0].observations[0].steps[0].gains[i][0][1], 0)
            self.assertEqual(project.sessions[0].observations[0].steps[0].gains[i][1][0], 0)
            self.assertEqual(project.sessions[0].observations[0].steps[0].gains[i][1][1], 1)
            
    def test_spc_write(self):
        """Test writing a STEPPED Delay and Gain SDF file."""
        
        project = sdf.parseSDF(spcFile)
        out = project.render()
        
    def test_spectrometer(self):
        """Test parsing DR spectrometer configurations."""
        project = sdf.parseSDF(drxFile)
        
        # Good spectrometer settings
        for channels in (2, 4, 8, 16, 32, 64, 128, 256, 512, 1024, 2048, 4096, 8192):
            for ints in (384, 768, 1536, 3072, 6144, 12288, 24576, 49152, 98304, 196608):
                for mode in (None, '', 'XXYY', 'IV', 'IQUV'):
                    ## Method 1
                    project.sessions[0].spcSetup = [channels, ints]
                    if mode in (None, ''):
                        project.sessions[0].spcMetatag = mode
                    else:
                        project.sessions[0].spcMetatag = '{Stokes=%s}' % mode
                    self.assertTrue(project.validate())
                    
                    ## Method 2
                    project.sessions[0].setSpectrometerChannels(channels)
                    project.sessions[0].setSpectrometerIntegration(ints)
                    if mode in (None, ''):
                        project.sessions[0].setSpectrometerMetatag(mode)
                    else:
                        project.sessions[0].setSpectrometerMetatag('Stokes=%s' % mode)
                    self.assertEqual(project.sessions[0].spcSetup[0], channels)
                    self.assertEqual(project.sessions[0].spcSetup[1], ints)
                    self.assertEqual(project.sessions[0].spcMetatag, None if mode in (None, '') else '{Stokes=%s}' % mode)
                    self.assertTrue(project.validate())
                    
        # Bad channel count
        project.sessions[0].spcSetup = [31, 6144]
        self.assertFalse(project.validate())
        
        # Bad integration count
        project.sessions[0].spcSetup = [32, 6145]
        self.assertFalse(project.validate())
        
        # Unsupported mode
        for mode in ('XX', 'XY', 'YX', 'YY', 'XXXYYXYY', 'I', 'Q', 'U', 'V'):
            project.sessions[0].spcSetup = [32, 6144]
            project.sessions[0].spcMetatag = '{Stokes=%s}' % mode
            self.assertFalse(project.validate())
            
    ### TBF ###
    
    def test_tbf_parse(self):
        """Test reading in a TBW SDF file."""
        
        self.assertRaises(RuntimeError, sdf.parseSDF, tbfFile)
        
    ### Misc. ###
    
    def test_auto_update(self):
        """Test project auto-update on render."""
        
        # Part 1 - frequency and duration
        project = sdf.parseSDF(drxFile)
        project.sessions[0].observations[1].frequency1 = 75e6
        project.sessions[0].observations[1].duration = '00:01:31.000'

        fh = open(os.path.join(self.testPath, 'sdf.txt'), 'w')		
        fh.write(project.render())
        fh.close()
        
        project = sdf.parseSDF(os.path.join(self.testPath, 'sdf.txt'))
        self.assertEqual(project.sessions[0].observations[1].freq1, 1643482384)
        self.assertEqual(project.sessions[0].observations[1].dur, 91000)
        
        # Part 2 - frequency and duration (timedelta)
        project = sdf.parseSDF(drxFile)
        project.sessions[0].observations[1].frequency1 = 75e6
        project.sessions[0].observations[1].duration = timedelta(minutes=1, seconds=31, microseconds=1000)

        fh = open(os.path.join(self.testPath, 'sdf.txt'), 'w')		
        fh.write(project.render())
        fh.close()
        
        project = sdf.parseSDF(os.path.join(self.testPath, 'sdf.txt'))
        self.assertEqual(project.sessions[0].observations[1].freq1, 1643482384)
        self.assertEqual(project.sessions[0].observations[1].dur, 91001)
        
        # Part 3 - frequency and start time
        project = sdf.parseSDF(drxFile)
        project.sessions[0].observations[1].frequency2 = 75e6
        project.sessions[0].observations[1].start = "MST 2011 Feb 23 17:00:15"

        fh = open(os.path.join(self.testPath, 'sdf.txt'), 'w')		
        fh.write(project.render())
        fh.close()
        
        project = sdf.parseSDF(os.path.join(self.testPath, 'sdf.txt'))
        self.assertEqual(project.sessions[0].observations[1].freq2, 1643482384)
        self.assertEqual(project.sessions[0].observations[1].mjd,  55616)
        self.assertEqual(project.sessions[0].observations[1].mpm,  15000)
        
        # Part 4 - frequency and start time (timedelta)
        project = sdf.parseSDF(drxFile)
        _MST = pytz.timezone('US/Mountain')
        project.sessions[0].observations[1].frequency2 = 75e6
        project.sessions[0].observations[1].start = _MST.localize(datetime(2011, 2, 23, 17, 00, 30, 1000))

        fh = open(os.path.join(self.testPath, 'sdf.txt'), 'w')		
        fh.write(project.render())
        fh.close()
        
        project = sdf.parseSDF(os.path.join(self.testPath, 'sdf.txt'))
        self.assertEqual(project.sessions[0].observations[1].freq2, 1643482384)
        self.assertEqual(project.sessions[0].observations[1].mjd,  55616)
        self.assertEqual(project.sessions[0].observations[1].mpm,  30001)
        
    def test_set_station(self):
        """Test the set stations functionlity."""
        
        project = sdf.parseSDF(drxFile)
        project.sessions[0].setStation(lwa1)
        self.assertTrue(project.validate())
        
        self.assertRaises(RuntimeError, project.sessions[0].setStation, lwasv)
        
    def test_is_valid(self):
        """Test whether or not isValid works."""
        
        self.assertTrue(sdf.isValid(tbwFile))
        self.assertTrue(sdf.isValid(tbnFile))
        self.assertTrue(sdf.isValid(drxFile))
        self.assertTrue(sdf.isValid(solFile))
        self.assertTrue(sdf.isValid(jovFile))
        self.assertTrue(sdf.isValid(stpFile))
        self.assertTrue(sdf.isValid(spcFile))
        
    def test_is_not_valid(self):
        """Test whether or not isValid works on LWA-SV and IDf files."""
        
        self.assertFalse(sdf.isValid(tbfFile))
        self.assertFalse(sdf.isValid(idfFile))
        
    def test_username(self):
        """Test setting auto-copy parameters."""
        
        project = sdf.parseSDF(drxFile)
        project.sessions[0].setDataReturnMethod('UCF')
        project.sessions[0].setUCFUsername('jdowell')
        out = project.render()
        
        self.assertTrue(out.find('Requested data return method is UCF') > 0)
        self.assertTrue(out.find('ucfuser:jdowell') > 0)
        
        fh = open(os.path.join(self.testPath, 'sdf.txt'), 'w')		
        fh.write(out)
        fh.close()
        
        project = sdf.parseSDF(os.path.join(self.testPath, 'sdf.txt'))
        out = project.render()
        
        self.assertTrue(out.find('Requested data return method is UCF') > 0)
        self.assertTrue(out.find('ucfuser:jdowell') > 0)
        
    def tearDown(self):
        """Remove the test path directory and its contents"""

        tempFiles = os.listdir(self.testPath)
        for tempFile in tempFiles:
            os.unlink(os.path.join(self.testPath, tempFile))
        os.rmdir(self.testPath)
        self.testPath = None


class sdf_test_suite(unittest.TestSuite):
    """A unittest.TestSuite class which contains all of the lsl.common.sdf units 
    tests."""
    
    def __init__(self):
        unittest.TestSuite.__init__(self)
        
        loader = unittest.TestLoader()
        self.addTests(loader.loadTestsFromTestCase(sdf_tests)) 


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-

# Python3 compatiability
import sys
if sys.version_info > (3,):
    xrange = range
    from functools import cmp_to_key
    
"""
Module for writing spectrometer output to a SDFITS file.  The SDFITS created by this 
modulefiles closely follow the Parkes variant of the SDFITS convention 
(see [http://fits.gsfc.nasa.gov/registry/sdfits.html]).  The main differences between 
the two is that the LWA SDFITS do not contain calbration or weather information.  This,
however, should not stop the files from being loaded into CASA with the ATNF Spectral 
Analysis Package (ASAP).

.. versionchanged:: 0.5.0
    The classes and functions defined in this module are based heavily off 
    the :mod:`lsl.writer.fitsidi` writer.
"""

import os
import gc
import re
import numpy
import pyfits
import warnings
from datetime import datetime

from lsl import astro
from lsl.common.stations import lwa1

__version__ = '0.5'
__revision__ = '$Rev: 2674 $'
__all__ = ['SD', '__version__', '__revision__', '__all__']


SDVersion = (1, 6)

StokesCodes = { 'I':  1,  'Q': 2,   'U':  3,  'V':  4, 
            'RR': -1, 'LL': -2, 'RL': -3, 'LR': -4, 
            'XX': -5, 'YY': -6, 'XY': -7, 'YX': -8}

NumericStokes = { 1: 'I',   2: 'Q',   3: 'U',   4: 'V', 
            -1: 'RR', -2: 'LL', -3: 'RL', -4: 'RL', 
            -5: 'XX', -6: 'YY', -7: 'XY', -8: 'YX'}


class SD(object):
    """
    Class for storing spectrometer data and writing the data, along with array
    frequency setup, etc., to a SDFITS file that can be read into CASA via the
    sd.scantable() function.
    """
    
    class _Frequency:
        """
        Holds information about the frequency setup used in the file.
        """

        def __init__(self, offset, channelWidth, bandwidth):
            self.id = 1
            self.bandFreq = offset
            self.chWidth = channelWidth
            self.totalBW = bandwidth
            self.sideBand = 1
            self.baseBand = 0
    
    class _SpectrometerData(object):
        """
        Representns one spectrum for a given observation time.
        """
        
        def __init__(self, obsTime, intTime, dataDict, pol=StokesCodes['XX']):
            self.obsTime = obsTime
            self.intTime = intTime
            self.dataDict = dataDict
            self.pol = pol
        
        def time(self):
            return self.obsTime
    
    def parseRefTime(self, refTime):
        """
        Given a time as either a integer, float, string, or datetime object, 
        convert it to a string in the formation 'YYYY-MM-DDTHH:MM:SS'.
        """

        # Valid time string (modulo the 'T')
        timeRE = re.compile(r'\d{4}-\d{2}-\d{2}[ T]\d{2}:\d{2}:\d{2}(\.\d+)?')

        if type(refTime) in (int, long, float):
            refDateTime = datetime.utcfromtimestamp(refTime)
            refTime = refDateTime.strftime("%Y-%m-%dT%H:%M:%S")
        elif type(refTime) == datetime:
            refTime = refTime.strftime("%Y-%m-%dT%H:%M:%S")
        elif type(refTime) == str:
            # Make sure that the string times are of the correct format
            if re.match(timeRE, refTime) is None:
                raise RuntimeError("Malformed date/time provided: %s" % refTime)
            else:
                refTime = refTime.replace(' ', 'T', 1)
        else:
            raise RuntimeError("Unknown time format provided.")

        return refTime

    def refTime2AstroDate(self):
        """
        Convert a reference time string to an :class:`lsl.astro.date` object.
        """

        dateStr = self.refTime.replace('T', '-').replace(':', '-').split('-')
        return astro.date(int(dateStr[0]), int(dateStr[1]), int(dateStr[2]), int(dateStr[3]), int(dateStr[4]), float(dateStr[5]))

    def __init__(self, filename, refTime=0.0, verbose=False, memmap=None, clobber=False):
        """
        Initialize a new SDFITS object using a filename and a reference time 
        given in seconds since the UNIX 1970 ephem, a python datetime object, or a 
        string in the format of 'YYYY-MM-DDTHH:MM:SS'.
        
        .. versionchanged:: 1.1.2
            Added the 'memmap' and 'clobber' keywords to control if the file
            is memory mapped and whether or not to overwrite an existing file, 
            respectively.
        """

        # File-specific information
        self.filename = filename
        self.verbose = verbose

        # Observation-specific information
        self.site = lwa1
        self.refTime = self.parseRefTime(refTime)
        self.nChan = 0
        self.nStokes = 0
        self.refVal = 0
        self.refPix = 0
        self.channelWidth = 0

        # Parameters that store the meta-data and data
        self.freq = []
        self.stokes = []
        self.data = []

        # Misc.
        self.tSys = 250
        self.observer = 'UKNOWN'
        self.project = 'UNKNOWN'
        self.mode = 'UNKNOWN'
        

        # Open the file and get going
        if os.path.exists(filename):
            if clobber:
                os.unlink(filename)
            else:
                raise IOError("File '%s' already exists" % filename)
        self.FITS = pyfits.open(filename, mode='append', memmap=memmap)

    def setSite(self, site):
        """
        Set the TELESCOP keyword in the primary HDU using an :class:`lsl.common.stations.LWAStation`
        object.
        """

        self.site = site
        
    def setStokes(self, polList):
        """
        Given a list of Stokes parameters, update the object's parameters.
        """

        for pol in polList:
            if type(pol) == str:
                numericPol = StokesCodes[pol.upper()]
            else:
                numericPol = pol
                
            if numericPol not in self.stokes:
                self.stokes.append(numericPol)
                
        # Sort into order of 'XX', 'YY', 'XY', and 'YX' or 'I', 'Q', 'U', and 'V'
        self.stokes.sort()
        if self.stokes[0] < 0:
            self.stokes.reverse()

        self.nStokes = len(self.stokes)

    def setFrequency(self, freq):
        """
        Given a numpy array of frequencies, set the relevant common observation
        parameters and add an entry to the self.freq list.
        """

        self.nChan = len(freq)
        self.refVal = freq[0]
        self.refPix = 1
        self.channelWidth = numpy.abs(freq[1] - freq[0])
        totalWidth = numpy.abs(freq[-1] - freq[0])

        freqSetup = self._Frequency(0.0, self.channelWidth, totalWidth)
        self.freq.append(freqSetup)
        
    def setObserver(self, observer, project='UNKNOWN', mode='UNKNOWN'):
        """
        Set the observer name, project, and observation mode (if given) to the 
        self.observer, self.project, and self.mode attributes, respectively.
        """
        
        self.observer = observer
        self.project = project
        self.mode = mode
        
    def addComment(self, comment):
        """
        Add a comment to data.
        
        .. versionadded:: 1.2.4
        """
        
        try:
            self._comments.append( comment )
        except AttributeError:
            self._comments = [comment,]
            
    def addHistory(self, history):
        """
        Add a history entry to the data.
        
        .. versionadded:: 1.2.4
        """
        
        try:
            self._history.append( history )
        except AttributeError:
            self._history = [history,]
            
    def addDataSet(self, obsTime, intTime, beam, data, pol='XX'):
        """
        Create a SpectrometerData object to store a collection of spectra.
        """
        
        if type(pol) == str:
            numericPol = StokesCodes[pol.upper()]
        else:
            numericPol = pol
        
        dataDict = {}
        for i,b in enumerate(beam):
            dataDict[b] = data[i,:]

        self.data.append( self._SpectrometerData(obsTime, intTime, dataDict, pol=numericPol) )

    def write(self):
        """
        Fill in the SDFITS file will all of the tables in the correct order.
        """
        
        def __sortData(x, y):
            """
            Function to sort the self.data list in order of time and then 
            polarization code.
            """
            
            xID = x.obsTime*10000000 + abs(x.pol)
            yID = y.obsTime*10000000 + abs(y.pol)
            
            if xID > yID:
                return 1
            elif xID < yID:
                return -1
            else:
                return 0
                
        # Sort the data set
        try:
            self.data.sort(cmp=__sortData)
        except TypeError:
            self.data.sort(key=cmp_to_key(__sortData))
            
        self.__writePrimary()
        self.__writeData()
        
        # Clear out the data section
        del(self.data[:])
        gc.collect()

    def close(self):
        """
        Close out the file.
        """

        self.FITS.flush()
        self.FITS.close()

    def __writePrimary(self):
        """
        Write the primary HDU to file.
        """

        primary = pyfits.PrimaryHDU()
        
        primary.header['NAXIS'] = (0, 'indicates SD file')
        primary.header['EXTEND'] = (True, 'indicates SD file')
        ts = str(astro.get_date_from_sys())
        primary.header['DATE'] = (ts.split()[0], 'IDI file creation date')
        primary.header['ORIGIN'] = 'LSL SDFITS writer'
        primary.header['TELESCOP'] = (self.site.name, 'Telescope name')
        
        # Write the comments and history
        try:
            for comment in self._comments:
                primary.header['comment'] = comment
            del self._comments
        except AttributeError:
            pass
        primary.header['COMMENT'] = " FITS (Flexible Image Transport System) format is defined in 'Astronomy and Astrophysics', volume 376, page 359; bibcode: 2001A&A...376..359H"
        try:
            for hist in self._history:
                primary.header['history'] = hist
            del self._history
        except AttributeError:
            pass
            
        self.FITS.append(primary)
        self.FITS.flush()
        
    def __writeData(self):
        """
        Define the SINGLE DISH table.
        """
        
        scanList = []
        dateList = []
        timeList = []
        intTimeList = []
        beamList = []
        mList = []
        rawList = []
        scanCount = 1
        for i,dataSet in enumerate(self.data):
            if dataSet.pol == self.stokes[0]:
                tempMList = {}
                for stokes in self.stokes:
                    tempMList[stokes] = {}
        
            beams = list(dataSet.dataDict.keys())
            beams.sort()
            for b in beams:
                specData = dataSet.dataDict[b]
                
                # Load the data into a matrix
                tempMList[dataSet.pol][b] = specData.ravel()
                
                if dataSet.pol == self.stokes[0]:
                    # Observation date and time
                    utc = astro.taimjd_to_utcjd(dataSet.obsTime)
                    date = astro.get_date(utc)
                    date.hours = 0
                    date.minutes = 0
                    date.seconds = 0
                    utc0 = date.to_jd()
                        
                    scanList.append(scanCount)
                    dateList.append('%4i-%02i-%02i' % (date.years, date.months, date.days))
                    timeList.append((utc - utc0)*24*3600)
                    intTimeList.append(dataSet.intTime)
                    beamList.append(b.id)
                    rawList.append(b)
            
            if dataSet.pol == self.stokes[-1]:
                for b in rawList:
                    matrix = numpy.zeros((self.nStokes,self.nChan), dtype=numpy.float32)
                    for p in xrange(self.nStokes):
                        try:
                            matrix[p,:] = tempMList[self.stokes[p]][b]
                        except KeyError:
                            warnings.warn('Key mis-match %s %s' % (str(b), str(tempMList[self.stokes[p]].keys())), RuntimeWarning)
                            
                    mList.append(matrix.ravel())
                scanCount += 1
                rawList = []
        
        # Scan number
        c1  = pyfits.Column(name='SCAN', format='1I', 
                        array=numpy.array(scanList))
        ## Cycle
        #c2 = pyfits.Column(name='CYCLE', format='1J', 
                        #array=numpy.array([1 for s in scanList]))
        # DATE-OBS
        c3  = pyfits.Column(name='DATE-OBS', format='10A', 
                        array = numpy.array(dateList))
        # Time elapsed since 0h
        c4  = pyfits.Column(name='TIME', format='1D', unit = 's', 
                        array = numpy.array(timeList))
        # Integration time (seconds)
        c5  = pyfits.Column(name='EXPOSURE', format='1E', unit='s', 
                        array=numpy.array(intTimeList, dtype=numpy.float32))
        # Object name
        c6  = pyfits.Column(name='OBJECT', format='16A', 
                        array=numpy.array(['LWA_OBS' for s in scanList]))
        # Object position (deg and deg)
        c7  = pyfits.Column(name='OBJ-RA', format='1D', unit='deg', 
                        array=numpy.array([0.0 for s in scanList]))
        c8  = pyfits.Column(name='OBJ-DEC', format='1D', unit='deg', 
                        array=numpy.array([0.0 for s in scanList]))
        # Rest frequency (Hz)
        c9  = pyfits.Column(name='RESTFRQ', format='1D', unit='Hz', 
                        array=numpy.array([0.0 for s in scanList]))
        # Observation mode
        c10 = pyfits.Column(name='OBSMODE', format='16A', 
                        array=numpy.array([self.mode for s in scanList]))
        # Beam (tuning)
        c11 = pyfits.Column(name='BEAM', format='1I', 
                        array=numpy.array(beamList))
        # IF
        c12 = pyfits.Column(name='IF', format='1I', 
                        array=numpy.array([self.freq[0].id for s in scanList]))
        # Frequency resolution (Hz)
        c13 = pyfits.Column(name='FREQRES', format='1D', unit='Hz', 
                        array=numpy.array([self.freq[0].chWidth for s in scanList]))
        # Bandwidth of the system (Hz)
        c14 = pyfits.Column(name='BANDWID', format='1D', unit='Hz', 
                        array=numpy.array([self.freq[0].totalBW for s in scanList]))
        # Frequency axis - 1
        c15 = pyfits.Column(name='CRPIX1', format='1E',
                        array=numpy.array([self.refPix for s in scanList]))
        c16 = pyfits.Column(name='CRVAL1', format='1D', unit='Hz', 
                        array=numpy.array([self.refVal for s in scanList]))
        c17 = pyfits.Column(name='CDELT1', format='1D', unit='Hz', 
                        array=numpy.array([self.freq[0].chWidth for s in scanList]))
        c18 = pyfits.Column(name='CRVAL3', format='1D', unit='deg', 
                        array=numpy.array([0.0 for s in scanList]))
        # Dec. axis - 4
        c19 = pyfits.Column(name='CRVAL4', format='1D', unit='deg', 
                        array=numpy.array([0.0 for s in scanList]))
        ## Scan rate
        #c20 = pyfits.Column(name='SCANRATE', format='2E', unit='deg/s', 
                        #array=numpy.array([[0,0] for s in scanList]))
                        
        #
        # Calibration information (currently not implemented)
        #
        ## System temperature  *** UNKNOWN ***
        #c21 =  pyfits.Column(name='TSYS', format='2E', unit='K', 
                        #array=numpy.array([[self.tSys,self.tSys] for s in scanList]))
        ## CALFCTR *** UNKNOWN ***
        #c22 =  pyfits.Column(name='CALFCTR', format='2E', unit='K', 
                        #array=numpy.array([[1,1] for s in scanList]))
        
        # Data
        c23 = pyfits.Column(name='DATA', format='%iE' % (self.nStokes*self.nChan), unit='UNCALIB', 
                        array=numpy.array(mList))
                        
        #
        # Data masking table (currently not implemented)
        #
        # Flag table
        #c24 = pyfits.Column(name='FLAGGED', format='%iB' % (self.nStokes*self.nChan), 
                        #array=numpy.array([[0,]*self.nStokes*self.nChan for s in scanList]))
        
        #
        # Calibration information (currently not implemented)
        #
        ## TCAL *** UNKNOWN ***
        #c25 = pyfits.Column(name='TCAL', format='2E', unit='Jy', 
                        #array=numpy.array([[1,1] for s in scanList]))
        ## TCALTIME *** UNKNOWN ***
        #c26 = pyfits.Column(name='TCALTIME', format='16A', 
                        #array=numpy.array(['UNKNOWN' for s in scanList]))
        
        #
        # Pointing information (currently not implemented)
        #
        ## Azimuth *** UNKNOWN ***
        #c27 = pyfits.Column(name='AZIMUTH', format='1E', unit='deg', 
                        #array=numpy.array([0 for s in scanList]))
        ## Elevation *** UNKNOWN ***
        #c28 = pyfits.Column(name='ELEVATIO', format='1E', unit='deg', 
                        #array=numpy.array([0 for s in scanList]))
        ## Parallactic angle *** UNKNOWN ***
        #c29 = pyfits.Column(name='PARANGLE', format='1E', unit='deg', 
                        #array=numpy.array([0 for s in scanList]))
        
        #
        # Focusing information (currently not implemented and probably never will be)
        #
        ## FOCUSAXI *** NOT NEEDED ***
        #c30 = pyfits.Column(name='FOCUSAXI', format='1E', unit='m', 
                        #array=numpy.array([0 for s in scanList]))
        ## FOCUSTAN *** NOT NEEDED ***
        #c31 = pyfits.Column(name='FOCUSTAN', format='1E', unit='m', 
                        #array=numpy.array([0 for s in scanList]))
        ## FOCUSROT *** NOT NEEDED ***
        #c32 = pyfits.Column(name='FOCUSROT', format='1E', unit='deg', 
                        #array=numpy.array([0 for s in scanList]))
        
        #
        # Weather information (currently not implemented)
        #
        ## Ambient temperature *** UNKNOWN ***
        #c33 = pyfits.Column(name='TAMBIENT', format='1E', unit='C', 
                        #array=numpy.array([0 for s in scanList]))
        ## Air pressure *** UNKNOWN ***
        #c34 = pyfits.Column(name='PRESSURE', format='1E', unit='Pa', 
                        #array=numpy.array([0 for s in scanList]))
        ## Humidity *** UNKNOWN ***
        #c35 = pyfits.Column(name='HUMIDITY', format='1E', unit='%', 
                        #array=numpy.array([0 for s in scanList]))
        ## Wind speed *** UNKNOWN ***
        #c36 = pyfits.Column(name='WINDSPEE', format='1E', unit='m/s', 
                        #array=numpy.array([0 for s in scanList]))
        ## Wind direction *** UNKNOWN ***
        #c37 = pyfits.Column(name='WINDDIRE', format='1E', unit='deg', 
                        #array=numpy.array([0 for s in scanList]))
        
        # Gather together all of the needed columns and figure out which ones
        # store the data and flag tables.  This information is needed later to
        # set the appropriate TDIM keywords.
        cs = []
        dataIndex = 0
        #flagIndex = 0
        n = 1
        for i in xrange(1, 38):
            try:
                cs.append(eval('c%i' % i))
                if eval('c%i.name' %i) == 'DATA':
                    dataIndex = n
                #if eval('c%i.name' %i) == 'FLAGGED':
                    #flagIndex = n
                n += 1
            except NameError:
                pass
        colDefs = pyfits.ColDefs(cs)
        
        # Create the SINGLE DISH table and update its header
        sd = pyfits.new_table(colDefs)
        
        ## Single disk keywords - order seems to matter
        sd.header['EXTNAME'] = ('SINGLE DISH', 'SDFITS table name')
        sd.header['NMATRIX'] = 1
        sd.header['OBSERVER'] = (self.observer, 'Observer name(s)')
        sd.header['PROJID']   = (self.project, 'Project name')
        sd.header['TELESCOP'] = (self.site.name, 'Telescope name')
        x,y,z = self.site.getGeocentricLocation()
        sd.header['OBSGEO-X'] = (x, '[m] Antenna ECEF X-coordinate')
        sd.header['OBSGEO-Y'] = (y, '[m] Antenna ECEF Y-coordinate')
        sd.header['OBSGEO-Z'] = (z, '[m] Antenna ECEF Z-coordinate')
        
        sd.header['SPECSYS'] = ('LSRK', 'Doppler reference frame (transformed)')
        sd.header['SSYSOBS'] = ('TOPOCENT', 'Doppler reference frame of observation')
        sd.header['EQUINOX'] = (2000.0, 'Equinox of equatorial coordinates')
        sd.header['RADESYS'] = ('FK5', 'Equatorial coordinate system frame')
        
        ## Data and flag table dimensionality
        sd.header['TDIM%i' % dataIndex] = ('(%i,%i,1,1)' % (self.nChan, self.nStokes))
        #sd.header.set('TDIM%i' % flagIndex, '(%i,%i,1,1)' % (self.nChan, self.nStokes), after='TFORM%i' % flagIndex)
        
        ## Data and flag table axis descriptions
        ### Frequency
        sd.header['CTYPE1'] = ('FREQ', 'axis 1 is FREQ (frequency)')
        sd.header['CDELT1'] = self.freq[0].chWidth
        sd.header['CRPIX1'] = self.refPix
        sd.header['CRVAL1'] = self.refVal
        ### Stokes
        sd.header['CTYPE2'] = ('STOKES', 'axis 2 is STOKES axis (polarization)')
        if self.stokes[0] < 0:
            sd.header['CDELT2'] = -1.0
        else:
            sd.header['CDELT2'] = 1.0
        sd.header['CRPIX2'] = 1.0
        sd.header['CRVAL2'] = float(self.stokes[0])
        ### RA
        sd.header['CTYPE3'] = ('RA', 'axis 3 is RA axis (pointing)')
        sd.header['CRPIX3'] = 1.0
        sd.header['CDELT3'] = -1.0
        ### Dec
        sd.header['CTYPE4'] = ('DEC', 'axis 4 is Dec. axis (pointing)')
        sd.header['CRPIX4'] = 1.0
        sd.header['CDELT4'] = 1.0
        
        self.FITS.append(sd)
        self.FITS.flush()



# -*- coding: utf-8 -*-

"""lsl.sim - Simulate various types of LWA data.  The following follow DP
format writers are avaliable:
  * tbw
  * tbn
  * drx

In addition, there are two simulation modules to generate fake data sets::
  * dp  - generate DP-level data sets for basic signals and point source
  * vis - generate visibility data sets for use with the aipy module
"""





# -*- coding: utf-8 -*-

"""Unit test for lsl.common.metabundle"""

import os
import unittest

from lsl.common import metabundleADP
from lsl.common.paths import dataBuild as dataPath


__revision__ = "$Rev: 2394 $"
__version__  = "0.1"
__author__    = "Jayce Dowell"

mdbFile = os.path.join(dataPath, 'tests', 'metadata.tgz')
mdbFileOld0 = os.path.join(dataPath, 'tests', 'metadata-old-0.tgz')
mdbFileOld1 = os.path.join(dataPath, 'tests', 'metadata-old-1.tgz')
mdbFileADP = os.path.join(dataPath, 'tests', 'metadata-adp.tgz')


class metabundle_tests_adp(unittest.TestCase):
    """A unittest.TestCase collection of unit tests for the lsl.common.metabundle
    module."""
    
    def test_ss(self):
        """Test the session specification utilties."""
        
        ses = metabundleADP.getSessionSpec(mdbFileADP)
        obs = metabundleADP.getObservationSpec(mdbFileADP)
        
        # Check session start time
        self.assertEqual(ses['MJD'], 57774)
        self.assertEqual(ses['MPM'], 29970000)
        
        # Check the duration
        self.assertEqual(ses['Dur'], obs[0]['Dur'] + 10000)
        
        # Check the number of observations
        self.assertEqual(ses['nObs'], len(obs))
    
    def test_os(self):
        """Test the observation specification utilities."""
        
        obs1 = metabundleADP.getObservationSpec(mdbFileADP)
        obs2 = metabundleADP.getObservationSpec(mdbFileADP, selectObs=1)
        
        # Check if the right observation is returned
        self.assertEqual(obs1[0], obs2)
        
        # Check the mode
        self.assertEqual(obs2['Mode'], 1)
        
        # Check the time
        self.assertEqual(obs2['MJD'], 57774)
        self.assertEqual(obs2['MPM'], 29975000)
        
    def test_cs(self):
        """Test the command script utilities."""
        
        cmnds = metabundleADP.getCommandScript(mdbFileADP)
        
        # Check number of command
        self.assertEqual(len(cmnds), 50)
        
        # Check the first and last commands
        self.assertEqual(cmnds[ 0]['commandID'], 'NUL')
        self.assertEqual(cmnds[-2]['commandID'], 'STP')
        self.assertEqual(cmnds[-1]['commandID'], 'ESN')
        
        # Check the counds of DP BAM commands
        nBAM = 0
        for cmnd in cmnds:
            if cmnd['commandID'] == 'BAM':
                nBAM += 1
        self.assertEqual(nBAM, 40)
        
    def test_sm(self):
        """Test the session metadata utilties."""
        
        sm = metabundleADP.getSessionMetaData(mdbFileADP)
        
        # Make sure all of the observations are done
        self.assertEqual(len(sm.keys()), 1)
        
    def test_sdf(self):
        """Test building a SDF from a tarball."""
        
        sdf = metabundleADP.getSessionDefinition(mdbFileADP)
        
    def test_sdm(self):
        """Test the station dynamic MIB utilties."""
        
        sm = metabundleADP.getSDM(mdbFileADP)
        
    def test_metadata(self):
        """Test the observation metadata utility."""
        
        fileInfo = metabundleADP.getSessionMetaData(mdbFileADP)
        self.assertEqual(len(fileInfo.keys()), 1)
        
        # File tag
        self.assertEqual(fileInfo[1]['tag'], '057774_000770030')
        
        # DRSU barcode
        self.assertEqual(fileInfo[1]['barcode'], 'S10TCC13S0016')
        
    def test_aspconfig(self):
        """Test retrieving the ASP configuration."""
        
        # Beginning config.
        aspConfig = metabundleADP.getASPConfigurationSummary(mdbFileADP, which='beginning')
        self.assertEqual(aspConfig['filter'],  0)
        self.assertEqual(aspConfig['at1'],     6)
        self.assertEqual(aspConfig['at2'],     5)
        self.assertEqual(aspConfig['atsplit'],15)
        
        # End config.
        aspConfig = metabundleADP.getASPConfigurationSummary(mdbFileADP, which='End')
        self.assertEqual(aspConfig['filter'],  0)
        self.assertEqual(aspConfig['at1'],     6)
        self.assertEqual(aspConfig['at2'],     5)
        self.assertEqual(aspConfig['atsplit'],15)
        
        # Unknown code
        self.assertRaises(ValueError, metabundleADP.getASPConfigurationSummary, mdbFileADP, 'middle')
        
    def test_is_valid(self):
        """Test whether or not isValid works."""
        
        self.assertTrue(metabundleADP.isValid(mdbFileADP))
        
    def test_is_not_valid(self):
        """Test whether or not isValid works on LWA1 files."""
        
        self.assertFalse(metabundleADP.isValid(mdbFile))
        self.assertFalse(metabundleADP.isValid(mdbFileOld0))
        self.assertFalse(metabundleADP.isValid(mdbFileOld1))


class metabundle_adp_test_suite(unittest.TestSuite):
    """A unittest.TestSuite class which contains all of the lsl.common.metabundleADP
    module unit tests."""
    
    def __init__(self):
        unittest.TestSuite.__init__(self)
        
        loader = unittest.TestLoader()
        self.addTests(loader.loadTestsFromTestCase(metabundle_tests_adp))
        
if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Given a TBF file, plot the time averaged spectra for each digitizer input."""

import os
import sys
import math
import ephem
import numpy
import argparse

from lsl.reader import tbf, errors
from lsl.common import stations, metabundleADP
from lsl.astro import unix_to_utcjd, DJD_OFFSET
from lsl.misc import parser as aph

from matplotlib import pyplot as plt


def bestFreqUnits(freq):
    """Given a numpy array of frequencies in Hz, return a new array with the
    frequencies in the best units possible (kHz, MHz, etc.)."""
    
    # Figure out how large the data are
    scale = int(math.log10(freq.max()))
    if scale >= 9:
        divis = 1e9
        units = 'GHz'
    elif scale >= 6:
        divis = 1e6
        units = 'MHz'
    elif scale >= 3:
        divis = 1e3
        units = 'kHz'
    else:
        divis = 1
        units = 'Hz'
        
    # Convert the frequency
    newFreq = freq / divis
    
    # Return units and freq
    return (newFreq, units)


def main(args):
    # Setup the LWA station information
    if args.metadata is not None:
        try:
            station = stations.parseSSMIF(args.metadata)
        except ValueError:
            station = metabundleADP.getStation(args.metadata, ApplySDM=True)
    else:
        station = stations.lwasv
    antennas = station.getAntennas()
    
    fh = open(args.filename, 'rb')
    nFrames = os.path.getsize(args.filename) / tbf.FrameSize
    antpols = len(antennas)
    
    # Read in the first frame and get the date/time of the first sample 
    # of the frame.  This is needed to get the list of stands.
    junkFrame = tbf.readFrame(fh)
    fh.seek(0)
    beginDate = ephem.Date(unix_to_utcjd(junkFrame.getTime()) - DJD_OFFSET)
    
    # Figure out how many frames there are per observation and the number of
    # channels that are in the file
    nFramesPerObs = tbf.getFramesPerObs(fh)
    nChannels = tbf.getChannelCount(fh)
    nSamples = 7840
    
    # Figure out how many chunks we need to work with
    nChunks = nFrames / nFramesPerObs
    
    # Pre-load the channel mapper
    mapper = []
    for i in xrange(2*nFramesPerObs):
        cFrame = tbf.readFrame(fh)
        if cFrame.header.firstChan not in mapper:
            mapper.append( cFrame.header.firstChan )
    fh.seek(-2*nFramesPerObs*tbf.FrameSize, 1)
    mapper.sort()
    
    # Calculate the frequencies
    freq = numpy.zeros(nChannels)
    for i,c in enumerate(mapper):
        freq[i*12:i*12+12] = c + numpy.arange(12)
    freq *= 25e3
    
    # File summary
    print "Filename: %s" % args.filename
    print "Date of First Frame: %s" % str(beginDate)
    print "Frames per Observation: %i" % nFramesPerObs
    print "Channel Count: %i" % nChannels
    print "Frames: %i" % nFrames
    print "==="
    print "Chunks: %i" % nChunks
    
    spec = numpy.zeros((nChannels,256,2))
    norm = numpy.zeros_like(spec)
    for i in xrange(nChunks):
        # Inner loop that actually reads the frames into the data array
        for j in xrange(nFramesPerObs):
            # Read in the next frame and anticipate any problems that could occur
            try:
                cFrame = tbf.readFrame(fh)
            except errors.eofError:
                break
            except errors.syncError:
                print "WARNING: Mark 5C sync error on frame #%i" % (int(fh.tell())/tbf.FrameSize-1)
                continue
            if not cFrame.header.isTBF():
                continue
                
            firstChan = cFrame.header.firstChan
            
            # Figure out where to map the channel sequence to
            try:
                aStand = mapper.index(firstChan)
            except ValueError:
                mapper.append(firstChan)
                aStand = mapper.index(firstChan)
            
            # Actually load the data.
            spec[aStand*12:aStand*12+12,:,:] += numpy.abs(cFrame.data.fDomain)**2
            norm[aStand*12:aStand*12+12,:,:] += 1
            
    spec /= norm
    fh.close()
    
    # Reshape and transpose to get it in to a "normal" order
    spec.shape = (spec.shape[0], spec.shape[1]*spec.shape[2])
    spec = spec.T
    
    # Apply the cable loss corrections, if requested
    if False:
        for s in range(spec.shape[0]):
            currGain = antennas[s].cable.gain(freq)
            spec[s,:] /= currGain
                
    # Put the frequencies in the best units possible
    freq, units = bestFreqUnits(freq)
    
    # Deal with the `keep` options
    if args.keep == 'all':
        antpolsDisp = int(numpy.ceil(antpols/20))
        js = [i for i in xrange(antpols)]
    else:
        antpolsDisp = int(numpy.ceil(len(args.keep)*2/20))
        if antpolsDisp < 1:
            antpolsDisp = 1
            
        js = []
        for k in args.keep:
            for i,ant in enumerate(antennas):
                if ant.stand.id == k:
                    js.append(i)
                    
    nPlot = len(js)
    if nPlot < 16:
        if nPlot % 4 == 0 and nPlot != 4:
            figsY = 4
        else:
            figsY = 2
        figsX = int(numpy.ceil(1.0*nPlot/figsY))
    else:
        figsY = 4
        figsX = 4
    figsN = figsX*figsY
    for i in xrange(antpolsDisp):
        # Normal plotting
        fig = plt.figure()
        for k in xrange(i*figsN, i*figsN+figsN):
            try:
                j = js[k]
                currSpectra = numpy.squeeze( numpy.log10(spec[j,:])*10.0 )
            except IndexError:
                break
            ax = fig.add_subplot(figsX, figsY, (k%figsN)+1)
            ax.plot(freq, currSpectra, label='Stand: %i, Pol: %i (Dig: %i)' % (antennas[j].stand.id, antennas[j].pol, antennas[j].digitizer))
            
            ax.set_title('Stand: %i (%i); Dig: %i [%i]' % (antennas[j].stand.id, antennas[j].pol, antennas[j].digitizer, antennas[j].getStatus()))
            ax.set_xlabel('Frequency [%s]' % units)
            ax.set_ylabel('P.S.D. [dB/RBW]')
            ax.set_ylim([-10, 30])
            
        # Save spectra image if requested
        if args.output is not None:
            base, ext = os.path.splitext(args.output)
            outFigure = "%s-%02i%s" % (base, i+1, ext)
            fig.savefig(outFigure)
            
        plt.draw()
        
    print "RBW: %.4f %s" % ((freq[1]-freq[0]), units)
    plt.show()


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
            description='read in a TBF file and create a collection of time-averaged spectra', 
            formatter_class=argparse.ArgumentDefaultsHelpFormatter
            )
    parser.add_argument('filename', type=str, 
                        help='filename to process')
    parser.add_argument('-m', '--metadata', type=str, 
                        help='name of the SSMIF or metadata tarball file to use for mappings')
    parser.add_argument('-q', '--quiet', dest='verbose', action='store_false',
                        help='run %(prog)s in silent mode')
    parser.add_argument('-k', '--keep', type=aph.csv_int_list, default='all', 
                        help='only display the following comma-seperated list of stands')
    parser.add_argument('-o', '--output', type=str, 
                        help='output file name for spectra image')
    args = parser.parse_args()
    main(args)
    



# -*- coding: utf-8 -*-

# Python3 compatiability
from __future__ import division

"""Unit test for the lsl.sim.dp module."""

import os
import unittest
import numpy
import tempfile

import aipy

from lsl.sim import dp
from lsl.reader import tbw
from lsl.reader import tbn
from lsl.reader import drx
from lsl.common import dp as dp_common
from lsl.common import stations as lwa_common


__version__  = "0.3"
__revision__ = "$Rev: 2470 $"
__author__    = "Jayce Dowell"


class simdp_tests(unittest.TestCase):
    """A unittest.TestCase collection of unit tests for the lsl.sim.dp
    module."""

    testPath = None

    def setUp(self):
        """Turn off all numpy warnings and create the temporary file directory."""

        numpy.seterr(all='ignore')
        self.testPath = tempfile.mkdtemp(prefix='test-simdp-', suffix='.tmp')
        self.src = aipy.src.get_catalog(['cyg',])
        
    def test_basic_tbw(self):
        """Test building a basic TBW signal"""

        testFile = os.path.join(self.testPath, 'tbw.dat')

        fh = open(testFile, 'wb')
        dp.basicSignal(fh, numpy.array([1,2,3,4]), 3000, mode='TBW', bits=12, tStart=1000)
        fh.close()

        # Check file size
        fileSize = os.path.getsize(testFile)
        nSamples = fileSize // tbw.FrameSize
        self.assertEqual(nSamples, 3000*4)

        # Check the time of the first frame
        fh = open(testFile, 'rb')
        frame = tbw.readFrame(fh)
        fh.close()
        self.assertEqual(frame.data.timeTag, 1000*dp_common.fS)
        self.assertEqual(frame.header.secondsCount, int(frame.getTime()))

        # Check that the frames have the correct value of data bits
        self.assertEqual(frame.getDataBits(), 12)

    def test_basic_tbn(self):
        """Test building a basic TBN signal"""

        testFile = os.path.join(self.testPath, 'tbn.dat')

        fh = open(testFile, 'wb')
        dp.basicSignal(fh, numpy.array([1,2,3,4]), 2000, mode='TBN', filter=7, tStart=1000)
        fh.close()

        # Check the file size
        fileSize = os.path.getsize(testFile)
        nSamples = fileSize // tbn.FrameSize
        self.assertEqual(nSamples, 2000*4*2)

        # Check the time of the first frame
        fh = open(testFile, 'rb')
        frame = tbn.readFrame(fh)
        fh.close()
        self.assertEqual(frame.data.timeTag, 1000*dp_common.fS)

    def test_point_tbw(self):
        """Test building a point source TBW signal"""

        testFile = os.path.join(self.testPath, 'tbw.dat')
        
        station = lwa_common.lwa1
        antennas = station.getAntennas()

        fh = open(testFile, 'wb')
        dp.pointSource(fh, antennas[:8:2], self.src, 3, mode='TBW', bits=12, tStart=1000)
        fh.close()

        # Check file size
        fileSize = os.path.getsize(testFile)
        nSamples = fileSize // tbw.FrameSize
        self.assertEqual(nSamples, 3*4)

        # Check the time of the first frame
        fh = open(testFile, 'rb')
        frame = tbw.readFrame(fh)
        fh.close()
        self.assertEqual(frame.data.timeTag, 1000*dp_common.fS)
        self.assertEqual(frame.header.secondsCount, int(frame.getTime()))

        # Check that the frames have the correct value of data bits
        self.assertEqual(frame.getDataBits(), 12)

    def test_point_tbn(self):
        """Test building a point source TBN signal"""

        testFile = os.path.join(self.testPath, 'tbn.dat')
        
        station = lwa_common.lwa1
        antennas = station.getAntennas()

        fh = open(testFile, 'wb')
        dp.pointSource(fh, antennas[:8:2], self.src, 4, mode='TBN', filter=7, tStart=1000)
        fh.close()

        # Check the file size
        fileSize = os.path.getsize(testFile)
        nSamples = fileSize // tbn.FrameSize
        self.assertEqual(nSamples, 4*4*2)

        # Check the time of the first frame
        fh = open(testFile, 'rb')
        frame = tbn.readFrame(fh)
        fh.close()
        self.assertEqual(frame.data.timeTag, 1000*dp_common.fS)

    def test_basic_drx(self):
        """Test building a basic DRX signal"""

        testFile = os.path.join(self.testPath, 'drx.dat')

        fh = open(testFile, 'wb')
        dp.basicSignal(fh, numpy.array([1,2,3,4]), 10, mode='DRX', filter=6, nTuning=2, tStart=1000)
        fh.close()

        # Check the file size
        fileSize = os.path.getsize(testFile)
        nSamples = fileSize // drx.FrameSize
        self.assertEqual(nSamples, 10*4*2*2)

        # Check the file size
        fh = open(testFile, 'rb')
        frame = drx.readFrame(fh)
        fh.close()
        self.assertEqual(frame.data.timeTag, 1000*dp_common.fS)
        self.assertEqual(frame.header.frameCount, 0)
        self.assertEqual(frame.header.secondsCount, 0)

    def tearDown(self):
        """Remove the test path directory and its contents"""

        tempFiles = os.listdir(self.testPath)
        for tempFile in tempFiles:
            os.unlink(os.path.join(self.testPath, tempFile))
        os.rmdir(self.testPath)
        self.testPath = None


class  simdp_test_suite(unittest.TestSuite):
    """A unittest.TestSuite class which contains all of the lsl.sim.vis units 
    tests."""
    
    def __init__(self):
        unittest.TestSuite.__init__(self)
        
        loader = unittest.TestLoader()
        self.addTests(loader.loadTestsFromTestCase(simdp_tests)) 


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-

# Python3 compatiability
import sys
if sys.version_info > (3,):
    xrange = range
    
"""
Module to allow for post-acquisition delay-and-sum beamforming with integer 
sample delays for TBW time series data (intDelayAndSum) and phase-and-sum 
beamforming for TBN time series data (delayAndSum).
"""

import os
import sys
import aipy
import math
import numpy

from lsl.common.paths import data as dataPath
from lsl.common.constants import c as speedOfLight
from lsl.common import dp as dp_common


__version__ = '0.6'
__revision__ = '$Rev: 2469 $'
__all__ = ['BeamformingError', 'calcDelay', 'intDelayAndSum', 'intBeamShape', 'phaseAndSum', 'phaseBeamShape', 'circularize', '__version__', '__revision__', '__all__']

class BeamformingError(Exception):
    """
    Base class for all exceptions in this file.
    """
    
    def __init__(self, strerror, *args, **kwds):
        try:
            self.errno = kwds['errno']
        except KeyError:
            self.errno = '-1'
        self.strerror = strerror
        self.filename = None
        self.args = (self.errno, strerror)
        super(BeamformingError, self).__init__(*args, **kwds)
        
    def __str__(self):
        return "%s" % self.strerror


def __loadStandResponse(freq=49.0e6):
    """
    Create an aipy.amp.beam object that holds the response for a single 
    isolated stand.  The stand response is based on NEC4 models at a variety
    of frequencies within the LWA frequency range.
    """
    
    # Read in the spherical harmonic representation of the beam distributed with
    # LSL
    dd = numpy.load(os.path.join(dataPath, 'beam-shape.npz'))
    coeffs = dd['coeffs']
    try:
        dd.close()
    except AttributeError:
        pass
        
    # Calculate how many harmonics are stored in the data set and reorder the data
    # to AIPY's liking
    deg = coeffs.shape[0]-1
    lmax = int((math.sqrt(1+8*coeffs.shape[1])-3)/2)
    beamShapeDict = {}
    for i in range(deg+1):
        beamShapeDict[i] = numpy.squeeze(coeffs[-1-i,:])
        
    # Build the beam object and done
    return aipy.amp.BeamAlm(numpy.array([freq/1e9]), lmax=lmax, mmax=lmax, deg=deg, nside=128, coeffs=beamShapeDict)


def calcDelay(antennas, freq=49.0e6, azimuth=0.0, elevation=90.0):
    """
    Calculate the time delays for delay-and-sum beam forming a collection of 
    stands looking in at a particular azimuth and elevation (both in degrees).  
    A numpy array of the geometric + cable delays in seconds is returned.
    
    .. versionchanged:: 0.5.0
        Changed the computed array center to exclude stands #257 through #260
    """
    
    # Make sure the pointing coordinates make sense
    if elevation < 0 or elevation > 90:
        raise BeamformingError("Pointing elevation (%.2f deg) is out of range [0, 90]" % elevation)
    if azimuth < 0 or azimuth > 360:
        raise BeamformingError("Pointing azimuth (%.2f deg) is out of range [0, 360]" % azimuth)
        
    # Get the positions of the stands and compute the mean center of the array
    xyz = numpy.zeros((len(antennas),3))
    i = 0
    good = []
    for ant in antennas:
        if ant.stand.id <= 256:
            good.append(i)
            
        xyz[i,0] = ant.stand.x
        xyz[i,1] = ant.stand.y
        xyz[i,2] = ant.stand.z
        i += 1
        
    arrayX = xyz[good,0].mean()
    arrayY = xyz[good,1].mean()
    arrayZ = xyz[good,2].mean()
    
    # Build up a unit vector that points in the direction azimuth,elevation
    rAz = azimuth*numpy.pi/180.0
    rEl = elevation*numpy.pi/180.0
    source = numpy.array([numpy.cos(rEl)*numpy.sin(rAz), 
                        numpy.cos(rEl)*numpy.cos(rAz), 
                        numpy.sin(rEl)])
                        
    # Compute the stand positions relative to the average and loop over stands
    # to compute the time delays in seconds
    arrayXYZ = xyz - numpy.array([arrayX, arrayY, arrayZ])
    delays = numpy.zeros((len(antennas),))
    for i in list(range(len(antennas))):
        delays[i] = numpy.dot(source, arrayXYZ[i,:]) / speedOfLight
        
    # Get the cable delays for each stand and add that in as well
    for i in list(range(len(antennas))):
        delays[i] = antennas[i].cable.delay(freq) - delays[i]
        
    # Done
    return delays


def intDelayAndSum(antennas, data, sampleRate=dp_common.fS, freq=49e6, azimuth=0.0, elevation=90.0):
    """
    Given a list of antennas and a 2-D data stream with stands enumerated
    along the first axis and time series samples along the second axis, 
    delay and sum the data stream into one beam.  The delays applied are 
    integer sample delays.  Return a 1-D numpy array of the time series data 
    associated with the formed beam.
    
    .. note:
        This task is primarily intended for use with TBW data.  The time resolution of
        TBN data, even at the highest bandwidth, has an integer sample delay of zero
        samples for all pointings.
        
    .. note:
        "Bad" antennas (those with antenna.getStatus() != 33) are automatically
        excluded from the beam.
        
    .. note:
        In order for this output to be properly processed by :func:`lsl.correlate.fx.calcSpectra`,
        The data need to be convert to a 2-D array via::
            
            >>> beamData = intDelayAndSum(stands, data)
            >>> beamData.shape = (1,)+beamData.shape
            
    .. versionchanged:: 0.4.0
        Switched over to passing in Antenna instances generated by the
        :mod:`lsl.common.stations` module instead of a list of stand ID
        numbers.
    """
    
    # Get the stand delays and convert the delay times from seconds to samples
    delays = calcDelay(antennas, freq=freq, azimuth=azimuth, elevation=elevation)
    delays = numpy.round(delays*sampleRate).astype(numpy.int16)
    
    # Figure out the polarizations
    pols = numpy.array([ant.pol for ant in antennas])
    pol0 = numpy.where( pols == 0 )[0]
    pol1 = numpy.where( pols == 1 )[0]
    
    # Make the delays into something meaningful for the shifting of the data 
    # streams
    delays -= delays.min()
    
    # Delay and sum by looping over stands inside of looping over times
    output = numpy.zeros((2, (data.shape[1]-delays.max())), dtype=data.dtype)
    for s,p in zip(range(len(antennas)), pols):
        if antennas[s].getStatus() != 33:
            continue
            
        start = delays[s]
        stop = data.shape[1] - delays.max() + start
        output[p,:] += data[s,start:stop]
        
    # Check for empty polarization data.  Always return a 2-D array for the data
    if len(pol0) == 0:
        output = output[1,:]
        output.shape = (1,) + output.shape
    elif len(pol1) == 0:
        output = output[0,:]
        output.shape = (1,) + output.shape
    else:
        pass
        
    # Done
    return output


def __intBeepAndSweep(antennas, arrayXYZ, t, freq, azimuth, elevation, beamShape=1.0, sampleRate=dp_common.fS, direction=(0.0, 90.0)):
    """
    Worker function for intBeamShape that 'beep's (makes a simulated signals) and
    'sweep's (delays it appropriately).
    """
    
    # Convert from degrees to radian
    rAz = azimuth*numpy.pi/180.0
    rEl = elevation*numpy.pi/180.0
    
    # Unit vector for the currect on-sky location
    currPos = numpy.array([numpy.cos(rEl)*numpy.sin(rAz), 
                        numpy.cos(rEl)*numpy.cos(rAz), 
                        numpy.sin(rEl)])
    # Stand response in this direction
    currResponse = beamShape
    
    # Loop over stands to build the simulated singnals
    signals = currResponse * numpy.exp(2j*numpy.pi*freq*t).astype(numpy.complex64)
    signals.shape = (1,signals.size)
    signals = numpy.repeat(signals, len(antennas), axis=0)
    delays = numpy.zeros((len(antennas),1), dtype=numpy.complex64)
    for i in xrange(len(antennas)):
        currDelay = antennas[i].cable.delay(freq) - numpy.dot(currPos, arrayXYZ[i,:]) / speedOfLight
        delays[i,0] = numpy.exp(-2j*numpy.pi*freq*currDelay)
    signals *= delays
        
    # Beamform with delay-and-sum and store the RMS result
    beamHere = intDelayAndSum(antennas, signals, sampleRate=sampleRate, freq=freq, azimuth=direction[0], elevation=direction[1])
    
    # Reduce the array dimensions
    beamHere = beamHere[0,:]
    
    # Return
    sigHere = (numpy.abs(beamHere)**2).mean()
    return sigHere


def intBeamShape(antennas, sampleRate=dp_common.fS, freq=49e6, azimuth=0.0, elevation=90.0, progress=False, DisablePool=False):
    """
    Given a list of antennas, compute the on-sky response of the delay-and-sum
    scheme implemented in intDelayAndSum.  A 360x90 numpy array spanning azimuth
    and elevation is returned.
    
    .. versionchanged:: 0.4.0
        Switched over to passing in Antenna instances generated by the
        :mod:`lsl.common.stations` module instead of a list of stand ID
        numbers.
        
    .. versionchanged:: 0.4.2
        Allowed for multiple polarization data to be delayed-and-summed correctly 
        and insured that a 2-D array is always returned (pol(s) by samples)
    """
    
    # Get the stand delays and convert the delay times from seconds to samples
    delays = calcDelay(antennas, freq=freq, azimuth=azimuth, elevation=elevation)
    delays = numpy.round(delays*sampleRate).astype(numpy.int16)
    
    # Build up a base time array, load in the cable delays, and get the stand 
    # positions for geometric delay calculations.
    t = numpy.arange(0,1500)/sampleRate
    xyz = numpy.zeros((len(antennas),3))
    i = 0
    good = []
    for ant in antennas:
        if ant.stand.id <= 256:
            good.append(i)
            
        xyz[i,0] = ant.stand.x
        xyz[i,1] = ant.stand.y
        xyz[i,2] = ant.stand.z
        i += 1
        
    arrayX = xyz[good,0].mean()
    arrayY = xyz[good,1].mean()
    arrayZ = xyz[good,2].mean()
    arrayXYZ = xyz - numpy.array([arrayX, arrayY, arrayZ])
    
    # Load in the response of a single isolated stand
    standBeam = __loadStandResponse(freq)
    
    # The multiprocessing module allows for the creation of worker pools to help speed
    # things along.  If the processing module is found, use it.  Otherwise, set
    # the 'usePool' variable to false and run single threaded.
    try:
        from multiprocessing import Pool, cpu_count
        
        # To get results pack from the pool, you need to keep up with the workers.  
        # In addition, we need to keep up with which workers goes with which 
        # baseline since the workers are called asynchronously.  Thus, we need a 
        # taskList array to hold tuples of baseline ('count') and workers.
        taskPool = Pool(processes=cpu_count())
        taskList = []
        
        usePool = True
        progress = False
    except ImportError:
        usePool = False
        
    # Turn off the thread pool if we are explicitly told not to use it.
    if DisablePool:
        usePool = False
        
    # Build up the beam shape over all azimuths and elevations
    beamShape =  numpy.zeros((360,90))
    for az in list(range(360)):
        rAz = az*numpy.pi/180.0
        for el in list(range(90)):
            rEl = el*numpy.pi/180.0
            beamShape[az,el] = standBeam.response(aipy.coord.azalt2top(numpy.concatenate([[rAz], [rEl]])))[0][0]
            
    # Build the output array and loop over all azimuths and elevations
    output = numpy.zeros((360,90))
    for az in list(range(360)):
        rAz = az*numpy.pi/180.0
        for el in list(range(90)):
            rEl = el*numpy.pi/180.0
            
            # Display the progress meter if the `progress' keyword is set to True.  The
            # progress meter displays a `.' every 2% complete and the percentages every
            # 10%.  At 100%, `Done' is displayed.
            if progress:
                fracDone = (az*90+el) / 32400.0 * 100
                if fracDone % 10 == 0 and round(fracDone, 2) != 100:
                    sys.stdout.write("%i%%" % fracDone)
                elif round(fracDone, 2) == 100:
                    sys.stdout.write("Done\n")
                elif round(fracDone, 3) % 2 == 0:
                    sys.stdout.write(".")
                else:
                    pass
                sys.stdout.flush()
                
            if usePool:
                task = taskPool.apply_async(__intBeepAndSweep, args=(antennas, arrayXYZ, t, freq, az, el), kwds={'beamShape': beamShape[az,el], 'sampleRate': sampleRate, 'direction': (azimuth, elevation)})
                taskList.append((az,el,task))
            else:
                output[az,el] = __intBeepAndSweep(antennas, arrayXYZ, t, freq, az, el, beamShape=beamShape[az,el], sampleRate=sampleRate, direction=(azimuth, elevation))
                
    # If pooling... Close the pool so that it knows that no ones else is joining.
    # Then, join the workers together and wait on the last one to finish before
    # saving the results.
    if usePool:
        taskPool.close()
        taskPool.join()
        
        # This is where he taskList list comes in handy.  We now know who did what
        # when we unpack the various results
        for az,el,task in taskList:
            output[az,el] = task.get()
            
    # Done
    return output


def phaseAndSum(antennas, data, sampleRate=dp_common.fS, CentralFreq=49.0e6, azimuth=0.0, elevation=90.0):
    """
    Given a list of antennas and a data stream of the form stands x times, 
    delay and sum the data stream into one beam.  Return a 1-D numpy array 
    of the time series data associated with the formed beam.
    
    .. note:
        This task is intended to be used with TBN data streams.
        
    .. note:
        "Bad" antennas (those with antenna.getStatus() != 33) are automatically
        excluded from the beam.
    """
    
    # Get the stand delays in seconds
    delays = calcDelay(antennas, freq=CentralFreq, azimuth=azimuth, elevation=elevation)
    
    # Make the delays into something meaningful for the shifting of the data 
    # streams.  Then, get the beamforming coefficients (b^l_n (a la Steve's 
    # "Fun with TBN" memo).
    delays -= delays.min()
    bln = numpy.exp(2j*numpy.pi*CentralFreq*delays)
    
    # Figure out the polarizations
    pols = numpy.array([ant.pol for ant in antennas])
    stat = numpy.array([ant.getStatus() for ant in antennas])
    pol0 = numpy.where( (pols == 0) & (stat == 33) )[0]
    pol1 = numpy.where( (pols == 1) & (stat == 33) )[0]
    
    # Loop over stands to compute the formed beam
    output = numpy.zeros((2, data.shape[1]), dtype=numpy.complex64)
    if len(pol0):
        output[0,:] = numpy.dot(bln[pol0], data[pol0,:]) / len(pol0)
    if len(pol1):
        output[1,:] = numpy.dot(bln[pol1], data[pol1,:]) / len(pol1)
        
    # Check for empty polarization data.  Always return a 3-D array for the data
    if len(pol0) == 0:
        output = output[1,:]
        output.shape = (1,) + output.shape
    elif len(pol1) == 0:
        output = output[0,:]
        output.shape = (1,) + output.shape
    else:
        pass
        
    # Done
    return output


def __phaseBeepAndSweep(antennas, arrayXYZ, t, freq, azimuth, elevation, beamShape=1.0, sampleRate=dp_common.fS, direction=(0.0, 90.0)):
    """
    Worker function for phaseBeamShape that 'beep's (makes a simulated signals) and
    'sweep's (phases it appropriately).
    """
    
    # Convert from degrees to radian
    rAz = azimuth*numpy.pi/180.0
    rEl = elevation*numpy.pi/180.0
        
    # Unit vector for the current on-sky location
    currPos = numpy.array([numpy.cos(rEl)*numpy.sin(rAz), 
                    numpy.cos(rEl)*numpy.cos(rAz), 
                    numpy.sin(rEl)])
    # Stand response in this direction
    currResponse = beamShape
    
    # Loop over stands to build the simulated signals
    signals = currResponse * numpy.exp(2j*numpy.pi*freq*t).astype(numpy.complex64)
    signals.shape = (1,signals.size)
    signals = numpy.repeat(signals, len(antennas), axis=0)
    delays = numpy.zeros((len(antennas),1), dtype=numpy.complex64)
    for i in xrange(len(antennas)):
        currDelay = antennas[i].cable.delay(freq) - numpy.dot(currPos, arrayXYZ[i,:]) / speedOfLight
        delays[i,0] = numpy.exp(-2j*numpy.pi*freq*currDelay)
    signals *= delays
        
    # Beamform with delay-and-sum and store the RMS result
    beam = phaseAndSum(antennas, signals, sampleRate=sampleRate, CentralFreq=freq, 
                        azimuth=direction[0], elevation=direction[1])
                        
    # Return
    sigHere = (numpy.abs(beam)**2).mean()
    return sigHere


def phaseBeamShape(antennas, sampleRate=dp_common.fS, CentralFreq=49.0e6, azimuth=0.0, elevation=90.0, progress=False):
    """
    Given a list of antennas, compute the on-sky response of the delay-and-sum
    scheme implemented in intDelayAndSum.  A 360x90 numpy array spanning azimuth
    and elevation is returned.
    
    .. versionchanged:: 1.2.1
        Removed the 'DisablePool' keyword since recent optimztions to 
        the function have caused the multiprocessing.Pool feature to
        actually be slower than the signle-threaded version.
        
    .. versionchanged:: 0.4.0
        Switched over to passing in Antenna instances generated by the
        :mod:`lsl.common.stations` module instead of a list of stand ID
        numbers.
    """
    
    # Get the stand delays and convert the delay times from seconds to samples
    delays = calcDelay(antennas, freq=CentralFreq, azimuth=azimuth, elevation=elevation)
    delays = numpy.round(delays*sampleRate).astype(numpy.int16)
    
    # Build up a base time array, load in the cable delays, and get the stand 
    # positions for geometric delay calculations.
    t = numpy.arange(0,1500)/sampleRate
    xyz = numpy.zeros((len(antennas),3))
    i = 0
    good = []
    for ant in antennas:
        if ant.stand.id <= 256:
            good.append(i)
            
        xyz[i,0] = ant.stand.x
        xyz[i,1] = ant.stand.y
        xyz[i,2] = ant.stand.z
        i += 1
        
    arrayX = xyz[good,0].mean()
    arrayY = xyz[good,1].mean()
    arrayZ = xyz[good,2].mean()
    arrayXYZ = xyz - numpy.array([arrayX, arrayY, arrayZ])
    
    # Load in the response of a single isolated stand
    standBeam = __loadStandResponse(freq=CentralFreq)
    
    # Build up the beam shape over all azimuths and elevations
    beamShape =  numpy.zeros((360,90))
    for az in xrange(360):
        rAz = az*numpy.pi/180.0
        for el in xrange(90):
            rEl = el*numpy.pi/180.0
            beamShape[az,el] = standBeam.response(aipy.coord.azalt2top(numpy.concatenate([[rAz], [rEl]])))[0][0]
            
    # Build the output array and loop over all azimuths and elevations
    output = numpy.zeros((360,90))
    for az in xrange(360):
        rAz = az*numpy.pi/180.0
        for el in xrange(90):
            rEl = el*numpy.pi/180.0
            
            # Display the progress meter if the `progress' keyword is set to True.  The
            # progress meter displays a `.' every 2% complete and the percentages every
            # 10%.  At 100%, `Done' is displayed.
            if progress:
                fracDone = (az*90+el) / 32400.0 * 100
                if fracDone % 10 == 0 and round(fracDone, 2) != 100:
                    sys.stdout.write("%i%%" % fracDone)
                elif round(fracDone, 2) == 100:
                    sys.stdout.write("Done\n")
                elif round(fracDone, 3) % 2 == 0:
                    sys.stdout.write(".")
                else:
                    pass
                sys.stdout.flush()
                
            output[az,el] = __phaseBeepAndSweep(antennas, arrayXYZ, t, CentralFreq, az, el, beamShape=beamShape[az,el], sampleRate=sampleRate, direction=(azimuth, elevation))
            
    # Done
    return output


def circularize(x, y, iauConvention=True):
    """
    Given a 1-D Numpy array of X polarization timeseries data and Y 
    polarization timeseries data, generate the two circular polarization.
    Returns a two-element tuple of L and R.
    
    .. versionchanged:: 1.0.1
        Added the 'iauConvention' keyword to help define the convention 
        of positive Stokes V.  V = RCP - LCP.
        
    .. versionadded:: 1.0.0
    """
    
    lrSign = 1.0
    if not iauConvention:
        lrSign = -1.0
        
    # Compute the circular terms
    l = (x + lrSign*1.0j*y) / numpy.sqrt(2)
    r = (x - lrSign*1.0j*y) / numpy.sqrt(2)
    
    # Done
    return l, r



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
import sys
import aipy
import pytz
import numpy
import pyfits
import argparse

from lsl import astro
from lsl.sim import vis as simVis
from lsl.imaging import utils, overlay
from lsl.writer.fitsidi import NumericStokes
from lsl.misc import parser as aph

import matplotlib.pyplot as plt
from matplotlib.ticker import NullFormatter


MST = pytz.timezone('US/Mountain')
UTC = pytz.UTC


NPIX_SIDE = 350


def main(args):
    filename = args.filename
    
    idi = utils.CorrelatedData(filename)
    aa = idi.getAntennaArray()
    lo = idi.getObserver()
    
    nStand = len(idi.stands)
    nChan = len(idi.freq)
    freq = idi.freq
    
    print "Raw Stand Count: %i" % nStand
    print "Final Baseline Count: %i" % (nStand*(nStand-1)/2,)
    print "Spectra Coverage: %.3f to %.3f MHz in %i channels (%.2f kHz/channel)" % (freq[0]/1e6, freq[-1]/1e6, nChan, (freq[1] - freq[0])/1e3)
    try:
        print "Polarization Products: %s" % ' '.join([NumericStokes[p] for p in idi.pols])
    except KeyError:
        # Catch for CASA MS that use a different numbering scheme
        NumericStokesMS = {1:'I', 2:'Q', 3:'U', 4:'V', 
                        9:'XX', 10:'XY', 11:'YX', 12:'YY'}
        print "Polarization Products: %s" % ' '.join([NumericStokesMS[p] for p in idi.pols])
        
    print "Reading in FITS IDI data"
    nSets = idi.integrationCount
    for set in range(1, nSets+1):
        if args.dataset != -1 and args.dataset != set:
            continue
            
        print "Set #%i of %i" % (set, nSets)
        dataDict = idi.getDataSet(set, uvMin=args.uv_min)
        
        # Build a list of unique JDs for the data
        pols = dataDict['bls'].keys()
        jdList = []
        for jd in dataDict['jd'][pols[0]]:
            if jd not in jdList:
                jdList.append(jd)
                
        # Find the LST
        lo.date = jdList[0] - astro.DJD_OFFSET
        utc = str(lo.date)
        lst = str(lo.sidereal_time())
        
        # Pull out the right channels
        toWork = numpy.where( (freq >= args.freq_start) & (freq <= args.freq_stop) )[0]
        if len(toWork) == 0:
            raise RuntimeError("Cannot find data between %.2f and %.2f MHz" % (args.freq_start/1e6, args.freq_stop/1e6))
            
        # Integration report
        print "    Date Observed: %s" % utc
        print "    LST: %s" % lst
        print "    Selected Frequencies: %.3f to %.3f MHz" % (freq[toWork[0]]/1e6, freq[toWork[-1]]/1e6)
        
        # Prune out what needs to go
        if args.include != 'all' or args.exclude != 'none':
            print "    Processing include/exclude lists"
            
            ## Create an empty output data dictionary
            newDict = {}
            for key in dataDict.keys():
                if key in ['bls', 'uvw', 'vis', 'wgt', 'msk', 'jd']:
                    newDict[key] = {}
                    for pol in dataDict[key].keys():
                        newDict[key][pol] = []
                else:
                    newDict[key] = dataDict[key]
                    
            ## Fill it
            for pol in dataDict['bls'].keys():
                for b in xrange(len(dataDict['bls'][pol])):
                    a0,a1 = dataDict['bls'][pol][b]
                    
                    if args.include != 'all':
                        if idi.stands[a0] in args.include and idi.stands[a1] in args.include:
                            for key in ['bls', 'uvw', 'vis', 'wgt', 'msk', 'jd']:
                                newDict[key][pol].append( dataDict[key][pol][b] )
                        else:
                            continue
                            
                    if args.exclude != 'none':
                        if idi.stands[a0] in args.exclude or idi.stands[a1] in args.exclude:
                            continue
                        else:
                            for key in ['bls', 'uvw', 'vis', 'wgt', 'msk', 'jd']:
                                newDict[key][pol].append( dataDict[key][pol][b] )
                                
            ## Make the substitution so that we are working with the right data now
            dataDict = newDict
            
            ## Report
            for pol in dataDict['bls'].keys():
                print "        %s now has %i baselines" % (pol, len(dataDict['bls'][pol]))
                
        # Build up the images for each polarization
        print "    Gridding"
        img1 = None
        lbl1 = 'XX'
        for p in ('xx', 'rr', 'I'):
            try:
                img1 = utils.buildGriddedImage(dataDict, MapSize=NPIX_SIDE/2, MapRes=0.5, pol=p, chan=toWork)
                lbl1 = p.upper()
            except:
                pass
                
        img2 = None
        lbl2 = 'YY'
        for p in ('yy', 'll', 'Q'):
            try:
                img2 = utils.buildGriddedImage(dataDict, MapSize=NPIX_SIDE/2, MapRes=0.5, pol=p, chan=toWork)
                lbl2 = p.upper()
            except:
                pass
                
        img3 = None
        lbl3 = 'XY'
        for p in ('xy', 'rl', 'U'):
            try:
                img3 = utils.buildGriddedImage(dataDict, MapSize=NPIX_SIDE/2, MapRes=0.5, pol=p, chan=toWork)
                lbl3 = p.upper()
            except:
                pass
                
        img4 = None
        lbl4 = 'YX'
        for p in ('yx', 'lr', 'V'):
            try:
                img4 = utils.buildGriddedImage(dataDict, MapSize=NPIX_SIDE/2, MapRes=0.5, pol=p, chan=toWork)
                lbl4 = p.upper()
            except:
                pass
                
        # Plots
        print "    Plotting"
        fig = plt.figure()
        ax1 = fig.add_subplot(2, 2, 1)
        ax2 = fig.add_subplot(2, 2, 2)
        ax3 = fig.add_subplot(2, 2, 3)
        ax4 = fig.add_subplot(2, 2, 4)
        for ax, img, pol in zip([ax1, ax2, ax3, ax4], [img1, img2, img3, img4], [lbl1, lbl2, lbl3, lbl4]):
            # Skip missing images
            if img is None:
                ax.text(0.5, 0.5, 'Not found in file', color='black', size=12, horizontalalignment='center')
                
                ax.xaxis.set_major_formatter( NullFormatter() )
                ax.yaxis.set_major_formatter( NullFormatter() )
                
                if not args.utc:
                    ax.set_title("%s @ %s LST" % (pol, lst))
                else:
                    ax.set_title("%s @ %s UTC" % (pol, utc))
                continue
                
            # Display the image and label with the polarization/LST
            cb = utils.plotGriddedImage(ax, img)
            fig.colorbar(cb, ax=ax)
            if not args.utc:
                ax.set_title("%s @ %s LST" % (pol, lst))
            else:
                ax.set_title("%s @ %s UTC" % (pol, utc))
                
            junk = img.image(center=(NPIX_SIDE/2,NPIX_SIDE/2))
            print "%s: image is %.4f to %.4f with mean %.4f" % (pol, junk.min(), junk.max(), junk.mean())
            
            # Turn off tick marks
            ax.xaxis.set_major_formatter( NullFormatter() )
            ax.yaxis.set_major_formatter( NullFormatter() )
            
            # Compute the positions of major sources and label the images
            overlay.sources(ax, aa, simVis.srcs, label=not args.no_labels)
            
            # Add in the horizon
            overlay.horizon(ax, aa)
            
            # Add lines of constant RA and dec.
            if not args.no_grid:
                if not args.topo:
                    overlay.graticleRADec(ax, aa)
                else:
                    overlay.graticleAzEl(ax, aa)
                    
        plt.show()
        
        if args.fits is not None:
            ## Loop over the images to build up the FITS file
            hdulist = [pyfits.PrimaryHDU(),]
            for img,pol in zip((img1,img2,img3,img4), (lbl1,lbl2,lbl3,lbl4)):
                if img is None:
                    continue
                    
                ### Create the HDU
                try:
                    hdu = pyfits.ImageHDU(data=img.image(center=(NPIX_SIDE/2,NPIX_SIDE/2)), name=pol)
                except AttributeError:
                    hdu = pyfits.ImageHDU(data=img, name=pol)
                    
                ### Add in the coordinate information
                hdu.header['EPOCH'] = 2000.0 + (jdList[0] - 2451545.0) / 365.25
                hdu.header['CTYPE1'] = 'RA---SIN'
                hdu.header['CRPIX1'] = NPIX_SIDE/2+1
                hdu.header['CDELT1'] = -360.0/NPIX_SIDE/numpy.pi
                hdu.header['CRVAL1'] = lo.sidereal_time()*180/numpy.pi
                hdu.header['CTYPE2'] = 'DEC--SIN'
                hdu.header['CRPIX2'] = NPIX_SIDE/2+1
                hdu.header['CDELT2'] = 360.0/NPIX_SIDE/numpy.pi
                hdu.header['CRVAL2'] = lo.lat*180/numpy.pi
                hdu.header['LONPOLE'] = 180.0
                hdu.header['LATPOLE'] = 90.0
                
                ### Add the HDU to the list
                hdulist.append(hdu)
                
            ## Save the FITS file to disk
            hdulist = pyfits.HDUList(hdulist)
            clobber = False
            if os.path.exists(args.fits):
                yn = raw_input("WARNING: '%s' exists, overwrite? [Y/n]" % args.fits)
                if yn not in ('n', 'N'):
                    clobber = True
            try:
                hdulist.writeto(args.fits, clobber=clobber)
            except IOError, e:
                print "WARNING: FITS image file not saved"
                
    print "...Done"


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description='create images from a FITS-IDI file', 
        epilog='NOTE:  If both -i/--include and -e/--exclude are specified the include list has priority.', 
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
        )
    parser.add_argument('filename', type=str, 
                        help='filename to plot')
    parser.add_argument('-1', '--freq-start', type=aph.frequency, default='10.0', 
                        help='first frequency to analyze in MHz')
    parser.add_argument('-2', '--freq-stop', type=aph.frequency, default='88.0', 
                        help='last frequency to analyze in MHz')
    parser.add_argument('-s', '--dataset', type=int, default=-1, 
                        help='data set to image')
    parser.add_argument('-m', '--uv-min', type=float, default=0.0, 
                        help='minimun baseline uvw length to include in lambda at the midpoint frequency')
    parser.add_argument('-i', '--include', type=aph.csv_int_list, default='all', 
                        help='comma seperated list of dipoles to include')
    parser.add_argument('-e', '--exclude', type=aph.csv_int_list, default='none', 
                        help='comma seperated list of dipoles to exclude')
    parser.add_argument('-t', '--topo', action='store_true', 
                        help='display an az/el grid instead of a RA/Dec grid')
    parser.add_argument('-u', '--utc', action='store_true', 
                        help='display the time as UTC, instead of LST')
    parser.add_argument('-n', '--no-labels', action='store_true', 
                        help='disable source and grid labels')
    parser.add_argument('-g', '--no-grid', action='store_true', 
                        help='disable the coordinate grid')
    parser.add_argument('-f', '--fits', type=str, 
                        help='save the images to the specified FITS image file')
    args = parser.parse_args()
    main(args)



# -*- coding: utf-8 -*-

# Python3 compatiability
import sys
if sys.version_info > (3,):
    xrange = range
    
"""
Module for taking a 2-D numpy array and turning it into an autostereogram
(http://en.wikipedia.org/wiki/Autostereogram).  This module also takes auto-
stereograms and returns the depth maps.

.. versionadded:: 0.5.3
"""

import numpy

__version__ = "0.1"
__revision__ = "$Rev: 2482 $"
__all__ = ['getASG', 'getDepthMap', '__version__', '__revision__', '__all__']


import warnings
warnings.warn("lsl.misc.autostereogram will be removed in LSL version 1.3.0", DeprecationWarning)


def getASG(data, patternWidth=100, patternPad=1, maxDepth=30):
    """
    Given a 2-D numpy array of data (image, elevation, etc.), convert the 
    data into a random dot autostereogram.  The options control the size of
    the pattern in the final image, the amount of padding around the image, 
    and how the input data is converted to a depth map.  The output is a 
    three-element tuple of the depth map, input random dot image, and auto-
    stereogram.  
    
    .. note::
        The output autostereogram is a 3-D numpy array with width, depth, and
        color channel.
    """
    
    # Autostereogram final size
    x = data.shape[0] + patternWidth*(2*patternPad)
    y = data.shape[1] + patternWidth*(2*patternPad)
    
    # Pattern to repeat (random colored dots)
    pattern = 255*numpy.random.rand(x, patternWidth, 3)
    pattern = pattern.astype(numpy.int16)
    
    # Number of tiles of the pattern needed to cover the autostereogram
    nTiles = 1.0*y/patternWidth
    
    # Base image and fill
    orig = numpy.zeros((x,y,3), dtype=numpy.int16)
    for i in xrange(int(nTiles)):
        start = i*pattern.shape[1]
        stop = start + pattern.shape[1]
        orig[:,start:stop,:] = pattern
    if nTiles > int(nTiles):
        start = (i+1)*pattern.shape[1]
        stop = orig.shape[1]
        orig[:,start:stop,:] = pattern[:,:(stop-start),:]
    
    # Compute the depth map and convert it to an integer
    dmap = data - data.min()
    dmap = numpy.round( float(maxDepth) * dmap/dmap.max() )
    dmap = dmap.astype(numpy.int16)

    # Shift it base image to make the autostereogram
    final = 1*orig
    for i in xrange(dmap.shape[0]):
        l = i + (patternWidth*patternPad)
        
        for j in xrange(dmap.shape[1]):
            m = j + (patternWidth*patternPad)
            d = patternWidth + dmap[i,j]
            
            final[l,m,:] = 1*final[l,m-d,:]
            
    # Return a three-element tuple of the map, base image, and autostereogram
    return dmap, orig, final


def getDepthMap(final):
    """
    Given an autostereogram image (width x heigth x color channel), determine the 
    depth map and return it.  This function uses the padding area around the input 
    image to determine the pattern width and then uses that width to look for 
    relative shifts.
    """
    
    # Find the pattern width by looking at the correlation of different 
    # columns of the first image channel
    l = numpy.arange(9, final.shape[1]/2)
    p = []
    for r in l:
        part1 = final[:,0:r,0].ravel().astype(numpy.float32)
        part2 = final[:,r:(2*r),0].ravel().astype(numpy.float32)
        p.append( numpy.correlate(part1, part2)[0] / r )
    p = numpy.array(p)
    p = numpy.abs(p)**2

    ## The pattern width occurs at the maximum
    patternWidth = l[numpy.where( p == p.max() )][0]
    
    # Find the depth map by 
    shift = numpy.zeros((final.shape[0]-2*patternWidth, final.shape[1]-2*patternWidth))
    for i in xrange(patternWidth, final.shape[0]-patternWidth):
        for j in xrange(patternWidth, final.shape[1]-patternWidth):
            d = numpy.zeros(patternWidth)
            for k in xrange(patternWidth):
                d[k] = ((final[i,j,:] - final[i,j-(k+patternWidth),:])**2).sum()
            d = numpy.sqrt(d)
            
            try:
                s = numpy.where( d == 0 )[0][0]
            except IndexError:
                s = 0
                
            shift[i-patternWidth,j-patternWidth] = s
            
    # Return the depth map
    return shift



# -*- coding: utf-8 -*-

"""
Python module to reading in data from TBF files.  This module defines the 
following classes for storing the TBF data found in a file:

Frame
  object that contains all data associated with a particular TBF frame.  The 
  primary consituents of each frame are:
    * FrameHeader - the TBF frame header object and
    * FrameData   - the TBF frame data object.  
Combined, these two objects contain all of the information found in the 
original TBF frame.

The functions defined in this module fall into two class:
  1. convert a frame in a file to a Frame object and
  2. describe the format of the data in the file.

For reading in data, use the readFrame function.  It takes a python file-
handle as an input and returns a fully-filled Frame object.

.. versionadded:: 1.2.0
"""

import copy
import numpy

from lsl.common import adp as adp_common
from lsl.reader._gofast import readTBF
from lsl.reader._gofast import syncError as gsyncError
from lsl.reader._gofast import eofError as geofError
from lsl.reader.errors import syncError, eofError

__version__ = '0.1'
__revision__ = '$Rev: 2782 $'
__all__ = ['FrameHeader', 'FrameData', 'Frame', 'readFrame', 'FrameSize', 'FrameChannelCount',
           'getFramesPerObs', 'getFirstFrameCount', 'getChannelCount', 'getFirstChannel', 
           '__version__', '__revision__', '__all__']

FrameSize = 6168
FrameChannelCount = 12


class FrameHeader(object):
    """
    Class that stores the information found in the header of a TBF 
    frame.  All three fields listed in the DP ICD version H are stored as 
    well as the original binary header data.
    """
    
    def __init__(self, id=None, frameCount=None, secondsCount=None, firstChan=None):
        self.id = id
        self.frameCount = frameCount
        self.secondsCount = secondsCount
        self.firstChan = firstChan
        
    def isTBF(self):
        """
        Function to check if the data is really TBF.  Returns True if the 
        data is TBF, false otherwise.
        """
        
        if self.id == 0x01:
            return True
        else:
            return False
            
    def getChannelFreqs(self):
        """
        Return a numpy.float32 array for the center frequencies, in Hz, of
        each channel in the data.
        """
        
        return (numpy.arange(12, dtype=numpy.float32)+self.firstChan) * adp_common.fC


class FrameData(object):
    """
    Class that stores the information found in the data section of a TBF
    frame.  Both fields listed in the DP ICD version H are stored.
    """
    
    def __init__(self, timeTag=None, fDomain=None):
        self.timeTag = timeTag
        self.fDomain = fDomain
        
    def getTime(self):
        """
        Function to convert the time tag from samples since the UNIX epoch
        (UTC 1970-01-01 00:00:00) to seconds since the UNIX epoch.
        """
        
        seconds = self.timeTag / adp_common.fS
        
        return seconds


class Frame(object):
    """
    Class that stores the information contained within a single TBF 
    frame.  It's properties are FrameHeader and FrameData objects.
    """
    
    def __init__(self, header=None, data=None):
        if header is None:
            self.header = FrameHeader()
        else:
            self.header = header
            
        if data is None:
            self.data = FrameData()
        else:
            self.data = data
            
        self.valid = True
        
    def isTBF(self):
        """
        Convenience wrapper for the Frame.FrameHeader.isTBF function.
        """
        
        return self.header.isTBF()
        
    def getChannelFreqs(self):
        """
        Convenience wrapper for the Frame.FrameHeader.getChannelFreqs function.
        """
        
        return self.header.getChannelFreqs()
        
    def getTime(self):
        """
        Convenience wrapper for the Frame.FrameData.getTime function.
        """
        
        return self.data.getTime()
        
    def __add__(self, y):
        """
        Add the data sections of two frames together or add a number 
        to every element in the data section.
        """
        
        newFrame = copy.deepcopy(self)
        newFrame += y
        return newFrame
        
    def __iadd__(self, y):
        """
        In-place add the data sections of two frames together or add 
        a number to every element in the data section.
        """
        
        try:
            self.data.fDomain += y.data.fDomain
        except AttributeError:
            self.data.fDomain += numpy.complex64(y)
        return self
        
    def __mul__(self, y):
        """
        Multiple the data sections of two frames together or multiply 
        a number to every element in the data section.
        """
        
        newFrame = copy.deepcopy(self)
        newFrame *= y
        return newFrame
            
    def __imul__(self, y):
        """
        In-place multiple the data sections of two frames together or 
        multiply a number to every element in the data section.
        """
        
        try:
            self.data.fDomain *= y.data.fDomain
        except AttributeError:
            self.data.fDomain *= numpy.complex64(y)
        return self
            
    def __eq__(self, y):
        """
        Check if the time tags of two frames are equal or if the time
        tag is equal to a particular value.
        """
        
        tX = self.data.timeTag
        try:
            tY = y.data.timeTag
        except AttributeError:
            tY = y
        
        if tX == tY:
            return True
        else:
            return False
            
    def __ne__(self, y):
        """
        Check if the time tags of two frames are not equal or if the time
        tag is not equal to a particular value.
        """
        
        tX = self.data.timeTag
        try:
            tY = y.data.timeTag
        except AttributeError:
            tY = y
        
        if tX != tY:
            return True
        else:
            return False
            
    def __gt__(self, y):
        """
        Check if the time tag of the first frame is greater than that of a
        second frame or if the time tag is greater than a particular value.
        """
        
        tX = self.data.timeTag
        try:
            tY = y.data.timeTag
        except AttributeError:
            tY = y
        
        if tX > tY:
            return True
        else:
            return False
            
    def __ge__(self, y):
        """
        Check if the time tag of the first frame is greater than or equal to 
        that of a second frame or if the time tag is greater than a particular 
        value.
        """
        
        tX = self.data.timeTag
        try:
            tY = y.data.timeTag
        except AttributeError:
            tY = y
        
        if tX >= tY:
            return True
        else:
            return False
            
    def __lt__(self, y):
        """
        Check if the time tag of the first frame is less than that of a
        second frame or if the time tag is greater than a particular value.
        """
        
        tX = self.data.timeTag
        try:
            tY = y.data.timeTag
        except AttributeError:
            tY = y
        
        if tX < tY:
            return True
        else:
            return False
            
    def __le__(self, y):
        """
        Check if the time tag of the first frame is less than or equal to 
        that of a second frame or if the time tag is greater than a particular 
        value.
        """
        
        tX = self.data.timeTag
        try:
            tY = y.data.timeTag
        except AttributeError:
            tY = y
        
        if tX <= tY:
            return True
        else:
            return False
            
    def __cmp__(self, y):
        """
        Compare two frames based on the time tags.  This is helpful for 
        sorting things.
        """
        
        tX = self.data.timeTag
        tY = y.data.timeTag
        if tY > tX:
            return -1
        elif tX > tY:
            return 1
        else:
            return 0


def readFrame(filehandle, Verbose=False):
    """
    Function to read in a single TBF frame (header+data) and store the 
    contents as a Frame object.
    """
    
    # New Go Fast! (TM) method
    try:
        newFrame = readTBF(filehandle, Frame())
    except gsyncError:
        mark = filehandle.tell() - FrameSize
        raise syncError(location=mark)
    except geofError:
        raise eofError
        
    return newFrame


def getFramesPerObs(filehandle):
    """
    Find out how many frames are present per time stamp by examining the 
    first 1000 TBF records.  Return the number of frames per observation.
    """
    
    # Save the current position in the file so we can return to that point
    fhStart = filehandle.tell()
    
    # Build up the list-of-lists that store the index of the first frequency
    # channel in each frame.
    channels = []
    for i in range(1000):
        try:
            cFrame = readFrame(filehandle)
            if not cFrame.isTBF():
                continue
        except eofError:
            break
        except syncError:
            continue
            
        chan = cFrame.header.firstChan
        if chan not in channels:
            channels.append( chan )
            
    # Return to the place in the file where we started
    filehandle.seek(fhStart)
    
    # Return the number of channels
    return len(channels)


def getFirstFrameCount(filehandle):
    """
    Find and return the lowest frame count encountered in a TBF file.
    """
    
    # Save the current position in the file so we can return to that point
    fhStart = filehandle.tell()
    
    # Find out how many frames there are per observation
    nFrames = getFramesPerObs(filehandle)
    
    firstFrameCount = 2**64-1
    freqs = []
    while len(freqs) < nFrames:
        cFrame = readFrame(filehandle)
        freq = cFrame.header.firstChan
            
        if freq not in freqs:
            freqs.append(freq)
        if cFrame.header.frameCount < firstFrameCount:
            firstFrameCount = cFrame.header.frameCount
            
    # Return to the place in the file where we started
    filehandle.seek(fhStart)
    
    # Return the lowest frame number found
    return firstFrameCount


def getChannelCount(filehandle):
    """
    Find out the total number of channels that are present by examining 
    the first 1000 TBF records.  Return the number of channels found.
    """
    
    # Find out how many frames there are per observation
    nFrames = getFramesPerObs(filehandle)
    
    # Convert to channels
    nChannels = nFrames * 12
    
    # Return the number of channels
    return nChannels


def getFirstChannel(filehandle, frequency=False):
    """
    Find and return the lowest frequency channel in a TBF file.  If the 
    `frequency` keyword is True the returned value is in Hz.
    """
    
    # Save the current position in the file so we can return to that point
    fhStart = filehandle.tell()
    
    # Find out how many frames there are per observation
    nFrames = getFramesPerObs(filehandle)
    
    # Find the lowest frequency channel
    freqs = []
    while len(freqs) < nFrames:
        cFrame = readFrame(filehandle)
        if frequency:
            freq = cFrame.getChannelFreqs()[0]
        else:
            freq = cFrame.header.firstChan
            
        if freq not in freqs:
            freqs.append(freq)
            
    # Return to the place in the file where we started
    filehandle.seek(fhStart)
    
    # Return the lowest frequency channel
    return min(freqs)



# -*- coding: utf-8 -*-

"""
Python module for creating creating, validating, and writing simulated 
DRX frames to a file.
"""

import numpy

from lsl.common.dp import fS
from lsl.reader import drx
from lsl.sim.errors import invalidBeam, invalidTune, invalidPol, invalidDataSize, invalidDataType

__version__ = '0.1'
__revision__ = '$Rev: 2467 $'
__all__ = ['SimFrame', 'frame2frame', '__version__', '__revision__', '__all__']


def frame2frame(drxFrame):
    """
    Convert a :class:`lsl.reader.drx.Frame` object to a raw DP DRX frame.
    """

    # The raw frame
    rawFrame = numpy.zeros(drx.FrameSize, dtype=numpy.uint8)

    # Part 1: The header
    ## Sync. words (0xDEC0DE5C)
    rawFrame[0] = 0xDE  # 222
    rawFrame[1] = 0xC0  # 192
    rawFrame[2] = 0xDE  # 222
    rawFrame[3] = 0x5C  #  92
    ## DRX ID
    rawFrame[4] = drxFrame.header.drxID
    ## Frame count
    rawFrame[5] = (drxFrame.header.frameCount>>16) & 255
    rawFrame[6] = (drxFrame.header.frameCount>>8) & 255
    rawFrame[7] = drxFrame.header.frameCount & 255
    ## Seconds count
    rawFrame[8] = (drxFrame.header.secondsCount>>24) & 255
    rawFrame[9] = (drxFrame.header.secondsCount>>16) & 255
    rawFrame[10] = (drxFrame.header.secondsCount>>8) & 255
    rawFrame[11] = drxFrame.header.secondsCount & 255
    ## Decimation
    rawFrame[12] = (drxFrame.header.decimation>>8) & 255
    rawFrame[13] = drxFrame.header.decimation & 255
    ## Time offset
    rawFrame[14] = (drxFrame.header.timeOffset>>8) & 255
    rawFrame[15] = drxFrame.header.timeOffset & 255

    # Part 2: The data
    ## Time tag
    rawFrame[16] = (drxFrame.data.timeTag>>56) & 255
    rawFrame[17] = (drxFrame.data.timeTag>>48) & 255
    rawFrame[18] = (drxFrame.data.timeTag>>40) & 255
    rawFrame[19] = (drxFrame.data.timeTag>>32) & 255
    rawFrame[20] = (drxFrame.data.timeTag>>24) & 255
    rawFrame[21] = (drxFrame.data.timeTag>>16) & 255
    rawFrame[22] = (drxFrame.data.timeTag>>8) & 255
    rawFrame[23] = drxFrame.data.timeTag & 255
    ## Flags
    rawFrame[24] = (drxFrame.data.flags>>56) & 255
    rawFrame[25] = (drxFrame.data.flags>>48) & 255
    rawFrame[26] = (drxFrame.data.flags>>40) & 255
    rawFrame[27] = (drxFrame.data.flags>>32) & 255
    rawFrame[28] = (drxFrame.data.flags>>24) & 255
    rawFrame[29] = (drxFrame.data.flags>>16) & 255
    rawFrame[30] = (drxFrame.data.flags>>8) & 255
    rawFrame[31] = drxFrame.data.flags & 255
    ## Data
    i = drxFrame.data.iq.real
    q = drxFrame.data.iq.imag
    ### Round, clip, and convert to unsigned integers
    i = i.round()
    i = i.clip(-8, 7)
    i = i.astype(numpy.int8)
    i += ((i & 8) << 1)
    q = q.round()
    q = q.clip(-8, 7)
    q = q.astype(numpy.int8)
    q += ((q & 8) << 1)
    
    rawFrame[32:] = (((i &  0xF) << 4) | (q & 0xF))
    
    return rawFrame


class SimFrame(drx.Frame):
    """
    drx.SimFrame extends the :class:`lsl.reader.drx.Frame` object to yield a method 
    for easily creating DP ICD-compliant raw DRX frames.  Frames created with
    this method can be written to a file via the methods writeRawFrame() function.
    """

    def __init__(self, beam=None, tune=None, pol=None, filterCode=None, timeOffset=None, frameCount=None, obsTime=None, flags=None, iq=None):
        """
        Given a list of parameters, build a drx.SimFrame object.  The parameters
        needed are:
        * beam id (>0 & <5)
        * tunning (1 or 2)
        * polarization (0 for x, or 1 for y)
        * which filter code the data corresponds to (>0 & <8)
        * what time offset in units of f_S to use
        * which frame number to create
        * observation time in samples at fS since the epoch
        * what flags are set on the data
        * 1-D numpy array representing the frame I/Q (complex) data
        
        Not all of these parameters are needed at initialization of the object and
        the values can be added later.

        .. versionchanged: 0.3.4
            obsTime now in samples at fS, not seconds
        """
        
        self.beam = beam
        self.tune = tune
        self.pol = pol
        self.filterCode = filterCode
        self.timeOffset = timeOffset
        self.frameCount = frameCount
        self.secondsCount = 0
        self.obsTime = obsTime
        self.flags = flags
        self.iq = iq
        super(SimFrame, self).__init__()
        
    def __update(self):
        """
        Private function to use the object's parameter values to build up 
        a drx.Frame-like object.
        """
        
        self.header.frameCount = 0*self.frameCount
        self.header.secondsCount = 0*long(self.obsTime / fS)
        self.header.decimation = int(fS / drx.filterCodes[self.filterCode])
        self.header.timeOffset = self.timeOffset
        self.header.drxID = (self.beam & 7) | ((self.tune & 7) << 3) | ((self.pol & 1) << 7)
        
        self.data.timeTag = self.obsTime
        self.data.flags = self.flags
        self.data.iq = self.iq
        
    def loadFrame(self, drxFrame):
        """
        Populate the a drx.SimFrame object with a pre-made frame.
        """
        
        self.header = drxFrame.header
        self.data = drxFrame.data

        inverseCodes = {}
        for code in drx.filterCodes:
            rate = drx.filterCodes[code]
            inverseCodes[int(rate)] = code
        
        # Back-fill the class' fields to make sure the object is consistent
        ## Header
        self.beam = self.header.parseID()[0]
        self.tune = self.header.parseID()[1]
        self.pol = self.header.parseID()[2]
        self.frameCount = self.header.frameCount
        self.secondsCount = self.header.secondsCount
        self.filterCode = inverseCodes[int(fS / self.header.decimation)]
        self.timeOffset = self.header.timeOffset
        ## Data
        self.obsTime = self.data.timeTag
        self.flags = self.data.flags
        self.iq = self.data.iq
    
    def isValid(self, raiseErrors=False):
        """
        Check if simulated DRX frame is valid or not.  Valid frames return 
        True and invalid frames False.  If the 'raiseErrors' keyword is set, 
        isValid raises an error when a problem is encountered.
        """

        # Make sure we have the latest values
        self.__update()

        # Is the time offset reasonable?
        if self.header.timeOffset >= fS:
            return False

        beam, tune, pol = self.parseID()
        # Is the beam number reasonable?
        if beam not in [1, 2, 3, 4]:
            if raiseErrors:
                raise invalidBeam()
            return False

        # Is the tunning number reasonable?
        if tune not in [1, 2]:
            if raiseErrors:
                raise invalidTune()
            return False

        # Is the polarization reasonable?
        if pol not in [0, 1]:
            if raiseErrors:
                raise invalidPol()
            return False

        # Is there data loaded into frame.data.iq?
        if self.data.iq is None:
            if raiseErrors:
                raise invalidDataSize()
            return False

        # Does the data length make sense?
        if self.data.iq.shape[0] != 4096:
            if raiseErrors:
                raise invalidDataSize()
            return False

        # Does the data type make sense?
        if self.data.iq.dtype.kind != 'c':
            if raiseErrors:
                raise invalidDataType()
            return False

        # If we made it this far, it's valid
        return True

    def createRawFrame(self):
        """
        Re-express a simulated DRX frame as a numpy array of unsigned 8-bit 
        integers.  Returns a numpy array if the frame is valid.  If the frame 
        is not ICD-compliant, a errors.baseSimError-type error is raised.
        """

        # Make sure we have the latest values
        self.__update()

        self.isValid(raiseErrors=True)
        return frame2frame(self)

    def writeRawFrame(self, fh):
        """
        Write a simulated DRX frame to a filehandle if the frame is valid.
        If the frame is not ICD-compliant, a errors.baseSimError-type error 
        is raised.
        """

        rawFrame = self.createRawFrame()
        rawFrame.tofile(fh)

    def __str__(self):
        if self.beam is None:
            return "Empty DRX SimFrame object"
        else:
            return "DRX SimFrame for beam %i, tunning %i, pol. %i @ time %i" % (self.beam, self.tune, self.pol, self.obsTime)
